"""Runtime Creator application discovery and dynamic app clients."""

from __future__ import annotations

import re
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, cast

from zoho.core.discovery_cache import DiscoveryDiskCache
from zoho.creator.models import CreatorResponse

if TYPE_CHECKING:
    from zoho.creator.client import CreatorClient


_INVALID_CLASS_CHAR_PATTERN = re.compile(r"[^0-9A-Za-z_]+")


class CreatorDynamicMetaClient:
    """Meta API wrapper bound to one Creator application."""

    def __init__(self, app_client: CreatorDynamicApplicationClient) -> None:
        self._app = app_client

    async def get_sections(
        self,
        *,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.meta.get_sections(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            headers=headers,
        )

    async def get_forms(
        self,
        *,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.meta.get_forms(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            headers=headers,
        )

    async def get_reports(
        self,
        *,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.meta.get_reports(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            headers=headers,
        )

    async def get_form_fields(
        self,
        *,
        form_link_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.meta.get_form_fields(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            form_link_name=form_link_name,
            headers=headers,
        )


class CreatorDynamicDataClient:
    """Data API wrapper bound to one Creator application."""

    def __init__(self, app_client: CreatorDynamicApplicationClient) -> None:
        self._app = app_client

    async def list_records(
        self,
        *,
        report_link_name: str,
        from_index: int | None = None,
        limit: int | None = None,
        criteria: str | None = None,
        record_cursor: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.data.list_records(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            from_index=from_index,
            limit=limit,
            criteria=criteria,
            record_cursor=record_cursor,
            headers=headers,
        )

    async def add_records(
        self,
        *,
        form_link_name: str,
        data: Mapping[str, Any] | Sequence[Mapping[str, Any]],
        skip_workflow: Sequence[str] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.data.add_records(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            form_link_name=form_link_name,
            data=data,
            skip_workflow=skip_workflow,
            headers=headers,
        )

    async def update_record(
        self,
        *,
        report_link_name: str,
        record_id: str | int,
        data: Mapping[str, Any],
        skip_workflow: Sequence[str] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.data.update_record(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            record_id=record_id,
            data=data,
            skip_workflow=skip_workflow,
            headers=headers,
        )

    async def delete_record(
        self,
        *,
        report_link_name: str,
        record_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.data.delete_record(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            record_id=record_id,
            headers=headers,
        )


class CreatorDynamicPublishClient:
    """Publish API wrapper bound to one Creator application."""

    def __init__(self, app_client: CreatorDynamicApplicationClient) -> None:
        self._app = app_client

    async def list_records(
        self,
        *,
        report_link_name: str,
        from_index: int | None = None,
        limit: int | None = None,
        criteria: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.publish.list_records(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            from_index=from_index,
            limit=limit,
            criteria=criteria,
            headers=headers,
        )

    async def add_records(
        self,
        *,
        form_link_name: str,
        data: Mapping[str, Any] | Sequence[Mapping[str, Any]],
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.publish.add_records(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            form_link_name=form_link_name,
            data=data,
            headers=headers,
        )

    async def update_record(
        self,
        *,
        report_link_name: str,
        record_id: str | int,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.publish.update_record(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            record_id=record_id,
            data=data,
            headers=headers,
        )

    async def delete_record(
        self,
        *,
        report_link_name: str,
        record_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        return await self._app._creator.publish.delete_record(
            account_owner_name=self._app.account_owner_name,
            app_link_name=self._app.app_link_name,
            report_link_name=report_link_name,
            record_id=record_id,
            headers=headers,
        )


class CreatorDynamicApplicationClient:
    """Runtime-bound Creator application client."""

    def __init__(
        self,
        creator_client: CreatorClient,
        *,
        account_owner_name: str,
        app_link_name: str,
        application_metadata: Mapping[str, Any] | None = None,
    ) -> None:
        self._creator = creator_client
        self._account_owner_name = account_owner_name
        self._app_link_name = app_link_name
        self._application_metadata = dict(application_metadata or {})
        self._meta: CreatorDynamicMetaClient | None = None
        self._data: CreatorDynamicDataClient | None = None
        self._publish: CreatorDynamicPublishClient | None = None

    @property
    def account_owner_name(self) -> str:
        return self._account_owner_name

    @property
    def app_link_name(self) -> str:
        return self._app_link_name

    @property
    def application_metadata(self) -> dict[str, Any]:
        return dict(self._application_metadata)

    @property
    def meta(self) -> CreatorDynamicMetaClient:
        if self._meta is None:
            self._meta = CreatorDynamicMetaClient(self)
        return self._meta

    @property
    def data(self) -> CreatorDynamicDataClient:
        if self._data is None:
            self._data = CreatorDynamicDataClient(self)
        return self._data

    @property
    def publish(self) -> CreatorDynamicPublishClient:
        if self._publish is None:
            self._publish = CreatorDynamicPublishClient(self)
        return self._publish


def _dynamic_application_class_name(app_link_name: str) -> str:
    clean = _INVALID_CLASS_CHAR_PATTERN.sub("_", app_link_name).strip("_")
    parts = [part for part in clean.split("_") if part]
    pascal_name = "".join(part[:1].upper() + part[1:] for part in parts)
    if not pascal_name:
        pascal_name = "Application"
    if pascal_name[0].isdigit():
        pascal_name = f"_{pascal_name}"
    return f"Creator{pascal_name}ApplicationClient"


class CreatorDynamicNamespace:
    """Dynamic Creator namespace with runtime application discovery."""

    def __init__(
        self,
        creator_client: CreatorClient,
        *,
        discovery_cache: DiscoveryDiskCache | None = None,
        discovery_cache_scope: str = "default:US:production",
    ) -> None:
        self._creator = creator_client
        self._clients: dict[str, CreatorDynamicApplicationClient] = {}
        self._dynamic_classes: dict[str, type[CreatorDynamicApplicationClient]] = {}
        self._application_metadata_by_key: dict[str, dict[str, Any]] = {}
        self._application_aliases: dict[str, str] = {}
        self._application_list: list[dict[str, Any]] = []
        self._last_loaded_scope: str | None = None
        self._discovery_cache = discovery_cache
        self._discovery_cache_scope = discovery_cache_scope

    async def list_applications(
        self,
        *,
        account_owner_name: str | None = None,
        use_cache: bool = True,
    ) -> list[dict[str, Any]]:
        scope = self._scope_for_workspace(account_owner_name)

        if use_cache and self._last_loaded_scope == scope and self._application_list:
            return list(self._application_list)

        if use_cache and self._discovery_cache is not None:
            cached = self._discovery_cache.load(
                product="creator",
                resource="applications",
                scope=scope,
            )
            if cached:
                self._ingest_applications(
                    cached, account_owner_hint=account_owner_name, scope=scope
                )
                return list(self._application_list)

        applications = (
            await self._creator.meta.list_applications_by_workspace(
                account_owner_name=account_owner_name,
            )
            if account_owner_name
            else await self._creator.meta.list_applications()
        )
        self._ingest_applications(applications, account_owner_hint=account_owner_name, scope=scope)

        if self._discovery_cache is not None and applications:
            self._discovery_cache.save(
                product="creator",
                resource="applications",
                scope=scope,
                entries=applications,
            )

        return list(self._application_list)

    async def has_application(
        self,
        name: str,
        *,
        account_owner_name: str | None = None,
        use_cache: bool = True,
    ) -> bool:
        await self.list_applications(account_owner_name=account_owner_name, use_cache=use_cache)
        return self._resolve_application_key(name) is not None

    async def get_application_client(
        self,
        name: str,
        *,
        account_owner_name: str | None = None,
        use_cache: bool = True,
    ) -> CreatorDynamicApplicationClient:
        await self.list_applications(account_owner_name=account_owner_name, use_cache=use_cache)
        canonical_key = self._resolve_application_key(name)
        if canonical_key is None:
            sample = ", ".join(sorted(self._application_aliases)[:8])
            message = f"Unknown Creator application: {name!r}"
            if sample:
                message = f"{message}. Known applications include: {sample}"
            raise KeyError(message)
        return self._client_for_application(canonical_key)

    async def precompile_applications(
        self,
        *,
        account_owner_name: str | None = None,
    ) -> list[dict[str, Any]]:
        """Force-refresh Creator applications and write them to disk cache."""

        return await self.list_applications(account_owner_name=account_owner_name, use_cache=False)

    def __getattr__(self, name: str) -> CreatorDynamicApplicationClient:
        if name.startswith("_"):
            raise AttributeError(f"{self.__class__.__name__} has no attribute {name!r}")
        canonical_key = self._resolve_application_key(name)
        if canonical_key is None:
            raise AttributeError(
                f"{self.__class__.__name__} has no dynamic application {name!r}; "
                "call list_applications() first."
            )
        return self._client_for_application(canonical_key)

    def _scope_for_workspace(self, account_owner_name: str | None) -> str:
        if account_owner_name is None:
            return self._discovery_cache_scope
        return f"{self._discovery_cache_scope}|workspace={account_owner_name}"

    @staticmethod
    def _application_key(account_owner_name: str, app_link_name: str) -> str:
        return f"{account_owner_name.lower()}:{app_link_name.lower()}"

    @staticmethod
    def _extract_application_identity(
        app: Mapping[str, Any],
        *,
        account_owner_hint: str | None,
    ) -> tuple[str | None, str | None]:
        app_link_name_raw = (
            app.get("link_name") or app.get("app_link_name") or app.get("application_link_name")
        )
        account_owner_raw = (
            app.get("workspace_name")
            or app.get("account_owner_name")
            or app.get("created_by")
            or account_owner_hint
        )
        app_link_name = (
            app_link_name_raw.strip()
            if isinstance(app_link_name_raw, str) and app_link_name_raw.strip()
            else None
        )
        account_owner_name = (
            account_owner_raw.strip()
            if isinstance(account_owner_raw, str) and account_owner_raw.strip()
            else None
        )
        return account_owner_name, app_link_name

    def _ingest_applications(
        self,
        applications: Sequence[Mapping[str, Any]],
        *,
        account_owner_hint: str | None,
        scope: str,
    ) -> None:
        metadata_by_key: dict[str, dict[str, Any]] = {}
        aliases: dict[str, str] = {}
        list_payload: list[dict[str, Any]] = []
        app_keys_by_link_name: dict[str, list[str]] = {}
        app_keys_by_attr_alias: dict[str, list[str]] = {}

        for app in applications:
            account_owner_name, app_link_name = self._extract_application_identity(
                app,
                account_owner_hint=account_owner_hint,
            )
            if account_owner_name is None or app_link_name is None:
                continue

            app_key = self._application_key(account_owner_name, app_link_name)
            metadata = dict(app)
            metadata.setdefault("account_owner_name", account_owner_name)
            metadata.setdefault("app_link_name", app_link_name)
            metadata_by_key[app_key] = metadata
            list_payload.append(metadata)

            aliases[app_key] = app_key
            aliases[f"{account_owner_name.lower()}.{app_link_name.lower()}"] = app_key
            app_keys_by_link_name.setdefault(app_link_name.lower(), []).append(app_key)
            attr_alias = app_link_name.lower().replace("-", "_")
            app_keys_by_attr_alias.setdefault(attr_alias, []).append(app_key)

        for app_link_name, app_keys in app_keys_by_link_name.items():
            if len(app_keys) == 1:
                aliases[app_link_name] = app_keys[0]
        for attr_alias, app_keys in app_keys_by_attr_alias.items():
            if len(app_keys) == 1:
                aliases[attr_alias] = app_keys[0]

        self._application_metadata_by_key = metadata_by_key
        self._application_aliases = aliases
        self._application_list = list_payload
        self._last_loaded_scope = scope

    def _resolve_application_key(self, name: str) -> str | None:
        name_key = name.lower()
        resolved = self._application_aliases.get(name_key)
        if resolved is not None:
            return resolved
        return self._application_aliases.get(name_key.replace("-", "_"))

    def _client_for_application(self, app_key: str) -> CreatorDynamicApplicationClient:
        existing = self._clients.get(app_key)
        if existing is not None:
            return existing

        metadata = self._application_metadata_by_key.get(app_key)
        if metadata is None:
            raise KeyError(f"Unknown Creator application key: {app_key!r}")

        account_owner_name = cast(str, metadata["account_owner_name"])
        app_link_name = cast(str, metadata["app_link_name"])

        dynamic_class = self._dynamic_classes.get(app_key)
        if dynamic_class is None:
            class_name = _dynamic_application_class_name(app_link_name)
            dynamic_class = cast(
                type[CreatorDynamicApplicationClient],
                type(
                    class_name,
                    (CreatorDynamicApplicationClient,),
                    {
                        "__doc__": (
                            f"Runtime Creator application client for {account_owner_name!r}/"
                            f"{app_link_name!r}."
                        )
                    },
                ),
            )
            self._dynamic_classes[app_key] = dynamic_class

        client = dynamic_class(
            self._creator,
            account_owner_name=account_owner_name,
            app_link_name=app_link_name,
            application_metadata=metadata,
        )
        self._clients[app_key] = client
        return client



"""Creator metadata operations."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.creator.models import CreatorResponse, parse_creator_response

if TYPE_CHECKING:
    from zoho.creator.client import CreatorClient


class CreatorMetaClient:
    """High-usage Creator metadata operations.

    Example:
        ```python
        sections = await zoho.creator.meta.get_sections(
            account_owner_name="owner",
            app_link_name="app",
        )
        ```
    """

    def __init__(self, creator_client: CreatorClient) -> None:
        self._creator = creator_client

    async def get_sections(
        self,
        *,
        account_owner_name: str,
        app_link_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        payload = await self._creator.request(
            "GET",
            f"/meta/{account_owner_name}/{app_link_name}/sections",
            headers=headers,
        )
        return parse_creator_response(payload)

    async def list_applications(
        self,
        *,
        headers: Mapping[str, str] | None = None,
    ) -> list[dict[str, Any]]:
        payload = await self._creator.request(
            "GET",
            "/creator/v2/meta/applications",
            headers=headers,
        )
        applications = payload.get("applications")
        return (
            [item for item in applications if isinstance(item, dict)]
            if isinstance(applications, list)
            else []
        )

    async def list_applications_by_workspace(
        self,
        *,
        account_owner_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> list[dict[str, Any]]:
        payload = await self._creator.request(
            "GET",
            f"/creator/v2/meta/{account_owner_name}/applications",
            headers=headers,
        )
        applications = payload.get("applications")
        return (
            [item for item in applications if isinstance(item, dict)]
            if isinstance(applications, list)
            else []
        )

    async def get_forms(
        self,
        *,
        account_owner_name: str,
        app_link_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        payload = await self._creator.request(
            "GET",
            f"/meta/{account_owner_name}/{app_link_name}/forms",
            headers=headers,
        )
        return parse_creator_response(payload)

    async def get_reports(
        self,
        *,
        account_owner_name: str,
        app_link_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        payload = await self._creator.request(
            "GET",
            f"/meta/{account_owner_name}/{app_link_name}/reports",
            headers=headers,
        )
        return parse_creator_response(payload)

    async def get_form_fields(
        self,
        *,
        account_owner_name: str,
        app_link_name: str,
        form_link_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> CreatorResponse:
        payload = await self._creator.request(
            "GET",
            f"/meta/{account_owner_name}/{app_link_name}/form/{form_link_name}/fields",
            headers=headers,
        )
        return parse_creator_response(payload)



"""Cliq threads APIs."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any

from zoho.cliq.models import CliqResponse, parse_cliq_response

if TYPE_CHECKING:
    from zoho.cliq.client import CliqClient


class CliqThreadsClient:
    """Thread APIs for Zoho Cliq."""

    def __init__(self, cliq_client: CliqClient) -> None:
        self._cliq = cliq_client

    async def post_message(
        self,
        *,
        thread_id: str,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "POST",
            f"/threads/{thread_id}/message",
            json=dict(data),
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def list_followers(
        self,
        *,
        thread_id: str,
        limit: int | None = None,
        next_token: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        params: dict[str, Any] = {}
        if limit is not None:
            params["limit"] = limit
        if next_token:
            params["next_token"] = next_token

        payload = await self._cliq.request(
            "GET",
            f"/threads/{thread_id}/followers",
            params=params,
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def add_followers(
        self,
        *,
        thread_id: str,
        user_ids: Sequence[str],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "POST",
            f"/threads/{thread_id}/followers",
            json={"user_ids": list(user_ids)},
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def remove_followers(
        self,
        *,
        thread_id: str,
        user_ids: Sequence[str],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "DELETE",
            f"/threads/{thread_id}/followers",
            json={"user_ids": list(user_ids)},
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def close(
        self,
        *,
        thread_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request("POST", f"/threads/{thread_id}/close", headers=headers)
        return parse_cliq_response(payload)

    async def reopen(
        self,
        *,
        thread_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request("POST", f"/threads/{thread_id}/reopen", headers=headers)
        return parse_cliq_response(payload)

    async def get_parent_message(
        self,
        *,
        thread_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "GET",
            f"/threads/{thread_id}/parentmessage",
            headers=headers,
        )
        return parse_cliq_response(payload)



"""Typer CLI helpers for Zoho OAuth and scope planning workflows."""

from __future__ import annotations

import json
import os
import re
import shlex
import sys
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any

import click
import httpx
import typer
from dotenv import load_dotenv

app = typer.Typer(
    help="Zoho authentication and scope helper commands.",
    add_completion=False,
    no_args_is_help=True,
)

_ACCOUNTS_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://accounts.zoho.com",
    "EU": "https://accounts.zoho.eu",
    "IN": "https://accounts.zoho.in",
    "AU": "https://accounts.zoho.com.au",
    "JP": "https://accounts.zoho.jp",
    "CA": "https://accounts.zohocloud.ca",
    "SA": "https://accounts.zoho.sa",
    "CN": "https://accounts.zoho.com.cn",
}

_REDACTED = "***REDACTED***"
_SENSITIVE_KEY_MARKERS = ("token", "secret", "cookie", "authorization", "password")
_SCOPE_TOKEN_PATTERN = re.compile(r"^[0-9a-zA-Z._ ]+$")
_PRODUCT_NAME_BY_ALIAS: dict[str, str] = {
    "crm": "CRM",
    "creator": "Creator",
    "projects": "Projects",
    "people": "People",
    "sheet": "Sheet",
    "workdrive": "WorkDrive",
    "cliq": "Cliq",
    "analytics": "Analytics",
    "writer": "Writer",
    "mail": "Mail",
}


@dataclass(frozen=True, slots=True)
class ScopeSpec:
    read: tuple[str, ...]
    write: tuple[str, ...]
    common: tuple[str, ...]


_SCOPE_CATALOG: dict[str, ScopeSpec] = {
    "CRM": ScopeSpec(
        read=(
            "ZohoCRM.modules.READ",
            "ZohoCRM.settings.READ",
            "ZohoCRM.users.READ",
            "ZohoCRM.org.READ",
        ),
        write=(
            "ZohoCRM.modules.ALL",
            "ZohoCRM.settings.ALL",
            "ZohoCRM.users.READ",
            "ZohoCRM.org.READ",
        ),
        common=(
            "ZohoCRM.modules.READ",
            "ZohoCRM.modules.ALL",
            "ZohoCRM.settings.READ",
            "ZohoCRM.settings.ALL",
            "ZohoCRM.users.READ",
            "ZohoCRM.org.READ",
            "ZohoCRM.coql.READ",
            "ZohoCRM.apis.READ",
        ),
    ),
    "Creator": ScopeSpec(
        read=(
            "ZohoCreator.report.READ",
            "ZohoCreator.meta.form.READ",
            "ZohoCreator.meta.application.READ",
        ),
        write=(
            "ZohoCreator.report.ALL",
            "ZohoCreator.form.CREATE",
            "ZohoCreator.meta.application.READ",
        ),
        common=(
            "ZohoCreator.report.READ",
            "ZohoCreator.report.CREATE",
            "ZohoCreator.report.UPDATE",
            "ZohoCreator.report.DELETE",
            "ZohoCreator.form.CREATE",
            "ZohoCreator.meta.form.READ",
            "ZohoCreator.meta.application.READ",
            "ZohoCreator.dashboard.READ",
        ),
    ),
    "Projects": ScopeSpec(
        read=(
            "ZohoProjects.portals.READ",
            "ZohoProjects.projects.READ",
            "ZohoProjects.tasks.READ",
        ),
        write=(
            "ZohoProjects.projects.ALL",
            "ZohoProjects.tasks.ALL",
            "ZohoProjects.tasklists.ALL",
        ),
        common=(
            "ZohoProjects.portals.READ",
            "ZohoProjects.projects.READ",
            "ZohoProjects.projects.ALL",
            "ZohoProjects.tasklists.READ",
            "ZohoProjects.tasks.READ",
            "ZohoProjects.tasks.UPDATE",
            "ZohoProjects.users.READ",
            "ZohoProjects.search.READ",
            "ZohoPC.files.CREATE",
        ),
    ),
    "People": ScopeSpec(
        read=(
            "ZOHOPEOPLE.forms.READ",
            "ZOHOPEOPLE.employee.ALL",
        ),
        write=(
            "ZOHOPEOPLE.forms.ALL",
            "ZOHOPEOPLE.employee.ALL",
        ),
        common=(
            "ZOHOPEOPLE.forms.READ",
            "ZOHOPEOPLE.forms.ALL",
            "ZOHOPEOPLE.employee.ALL",
            "ZOHOPEOPLE.attendance.ALL",
            "ZOHOPEOPLE.leave.READ",
            "ZOHOPEOPLE.timetracker.ALL",
        ),
    ),
    "Sheet": ScopeSpec(
        read=("ZohoSheet.dataAPI.READ",),
        write=(
            "ZohoSheet.dataAPI.READ",
            "ZohoSheet.dataAPI.UPDATE",
        ),
        common=(
            "ZohoSheet.dataAPI.READ",
            "ZohoSheet.dataAPI.UPDATE",
        ),
    ),
    "WorkDrive": ScopeSpec(
        read=(
            "WorkDrive.team.READ",
            "WorkDrive.files.READ",
        ),
        write=(
            "WorkDrive.team.ALL",
            "WorkDrive.files.ALL",
        ),
        common=(
            "WorkDrive.team.ALL",
            "WorkDrive.files.READ",
            "WorkDrive.files.CREATE",
            "WorkDrive.files.UPDATE",
            "WorkDrive.files.DELETE",
        ),
    ),
    "Cliq": ScopeSpec(
        read=(
            "ZohoCliq.Users.READ",
            "ZohoCliq.Chats.READ",
            "ZohoCliq.Channels.READ",
            "ZohoCliq.Messages.READ",
        ),
        write=(
            "ZohoCliq.Users.UPDATE",
            "ZohoCliq.Channels.UPDATE",
            "ZohoCliq.Messages.UPDATE",
            "ZohoCliq.Webhooks.CREATE",
        ),
        common=(
            "ZohoCliq.Users.READ",
            "ZohoCliq.Users.UPDATE",
            "ZohoCliq.Chats.READ",
            "ZohoCliq.Channels.READ",
            "ZohoCliq.Channels.UPDATE",
            "ZohoCliq.Messages.READ",
            "ZohoCliq.Messages.UPDATE",
            "ZohoCliq.Webhooks.CREATE",
        ),
    ),
    "Analytics": ScopeSpec(
        read=(
            "ZohoAnalytics.metadata.read",
            "ZohoAnalytics.data.read",
        ),
        write=(
            "ZohoAnalytics.metadata.read",
            "ZohoAnalytics.data.create",
            "ZohoAnalytics.data.update",
        ),
        common=(
            "ZohoAnalytics.metadata.read",
            "ZohoAnalytics.data.read",
            "ZohoAnalytics.data.create",
            "ZohoAnalytics.data.update",
            "ZohoAnalytics.data.delete",
            "ZohoAnalytics.usermanagement.read",
        ),
    ),
    "Writer": ScopeSpec(
        read=("ZohoWriter.documentEditor.ALL",),
        write=(
            "ZohoWriter.documentEditor.ALL",
            "ZohoWriter.merge.ALL",
            "ZohoPC.files.ALL",
            "WorkDrive.files.ALL",
        ),
        common=(
            "ZohoWriter.documentEditor.ALL",
            "ZohoWriter.merge.ALL",
            "ZohoPC.files.ALL",
            "WorkDrive.files.ALL",
            "WorkDrive.organization.ALL",
            "WorkDrive.workspace.ALL",
            "ZohoSign.documents.ALL",
        ),
    ),
    "Mail": ScopeSpec(
        read=(
            "ZohoMail.accounts.READ",
            "ZohoMail.folders.READ",
            "ZohoMail.messages.READ",
            "ZohoMail.threads.READ",
        ),
        write=(
            "ZohoMail.accounts.READ",
            "ZohoMail.folders.ALL",
            "ZohoMail.messages.ALL",
            "ZohoMail.threads.ALL",
        ),
        common=(
            "ZohoMail.accounts.READ",
            "ZohoMail.folders.READ",
            "ZohoMail.folders.ALL",
            "ZohoMail.messages.READ",
            "ZohoMail.messages.ALL",
            "ZohoMail.threads.READ",
            "ZohoMail.threads.ALL",
        ),
    ),
}


class AccessProfile(StrEnum):
    read = "read"
    write = "write"
    admin = "admin"


class OutputFormat(StrEnum):
    json = "json"
    text = "text"
    env = "env"


def _print_json(payload: Any) -> None:
    typer.echo(json.dumps(payload, indent=2, sort_keys=True))


def _require_credentials_file(path: str | None) -> Path:
    if not path:
        raise typer.BadParameter("ZOHO_CREDENTIALS_FILE is required for this command")
    credentials_path = Path(path).expanduser()
    if not credentials_path.is_file():
        raise typer.BadParameter(f"Credentials file does not exist: {credentials_path}")
    return credentials_path


def _load_credentials_env(credentials_file: str | None) -> Path:
    resolved = _require_credentials_file(credentials_file or os.getenv("ZOHO_CREDENTIALS_FILE"))
    load_dotenv(dotenv_path=resolved, override=False)
    return resolved


def _required_env(name: str) -> str:
    value = os.getenv(name)
    if not value:
        raise typer.BadParameter(f"Missing required environment variable: {name}")
    return value


def _sanitize_payload(value: Any, *, show_secrets: bool) -> Any:
    if show_secrets:
        return value

    if isinstance(value, dict):
        redacted: dict[str, Any] = {}
        for key, item in value.items():
            key_lower = key.lower()
            if any(marker in key_lower for marker in _SENSITIVE_KEY_MARKERS):
                redacted[key] = _REDACTED
            else:
                redacted[key] = _sanitize_payload(item, show_secrets=show_secrets)
        return redacted
    if isinstance(value, list):
        return [_sanitize_payload(item, show_secrets=show_secrets) for item in value]
    return value


def _dedupe(values: list[str]) -> list[str]:
    deduped: list[str] = []
    seen: set[str] = set()
    for value in values:
        if value and value not in seen:
            seen.add(value)
            deduped.append(value)
    return deduped


def _extract_scopes(*, scopes_csv: str | None, scope_items: list[str] | None) -> list[str]:
    scopes: list[str] = []
    if scopes_csv:
        scopes.extend(scope.strip() for scope in scopes_csv.split(",") if scope.strip())
    if scope_items:
        scopes.extend(scope.strip() for scope in scope_items if scope.strip())
    return _dedupe(scopes)


def _validate_scope_tokens(scopes: list[str]) -> list[str]:
    invalid = [
        scope for scope in scopes if (not scope) or (_SCOPE_TOKEN_PATTERN.fullmatch(scope) is None)
    ]
    if invalid:
        pretty = ", ".join(repr(scope) for scope in invalid)
        raise typer.BadParameter(
            "Invalid scope value(s): "
            + pretty
            + ". Allowed token characters: letters, numbers, dot, underscore, and spaces."
        )
    return scopes


def _resolve_accounts_domain(explicit_domain: str | None) -> str:
    if explicit_domain:
        return explicit_domain.rstrip("/")
    env_domain = os.getenv("ZOHO_ACCOUNTS_DOMAIN")
    if env_domain:
        return env_domain.rstrip("/")
    dc = os.getenv("ZOHO_DC", "US").upper()
    return _ACCOUNTS_DOMAIN_BY_DC.get(dc, _ACCOUNTS_DOMAIN_BY_DC["US"])


def _parse_extra_headers(values: list[str] | None) -> dict[str, str]:
    headers: dict[str, str] = {}
    for item in values or []:
        if "=" not in item:
            raise typer.BadParameter(f"Invalid --header format: {item!r}. Expected KEY=VALUE.")
        key, value = item.split("=", 1)
        key = key.strip()
        value = value.strip()
        if not key:
            raise typer.BadParameter(f"Invalid --header format: {item!r}. Header key is empty.")
        headers[key] = value
    return headers


def _http_response_to_json(response: httpx.Response) -> dict[str, Any]:
    try:
        body = response.json()
    except ValueError:
        return {
            "status_code": response.status_code,
            "text": response.text[:500],
        }
    if isinstance(body, dict):
        body.setdefault("status_code", response.status_code)
        return body
    return {
        "status_code": response.status_code,
        "data": body,
    }


def _build_grant_payload(
    *,
    scopes_csv: str | None,
    scope_items: list[str] | None,
    expiry: int,
    description: str,
    approved_orgs: str | None,
) -> dict[str, Any]:
    scopes = _validate_scope_tokens(_extract_scopes(scopes_csv=scopes_csv, scope_items=scope_items))
    if not scopes:
        raise typer.BadParameter("At least one scope is required (--scopes or --scope)")

    grant_payload: dict[str, Any] = {
        "scope": scopes,
        "expiry": expiry,
        "description": description,
    }
    if approved_orgs:
        grant_payload["approvedorgs"] = approved_orgs
    return {"granttoken": grant_payload}


def _grant_endpoint(api_console_domain: str, self_client_id: str) -> str:
    return f"{api_console_domain.rstrip('/')}/oauthapi/v1/selfclient/{self_client_id}/granttoken"


def _render_grant_curl_template(endpoint: str, payload: dict[str, Any]) -> str:
    payload_raw = json.dumps(payload, separators=(",", ":"), ensure_ascii=True)
    return "\n".join(
        [
            f"curl {shlex.quote(endpoint)} \\",
            "  -H 'Accept: */*' \\",
            "  -H 'Content-Type: application/json' \\",
            "  -H 'Cookie: <SESSION_COOKIE>' \\",
            "  -H 'x-zcsrf-token: <X_ZCSRF_TOKEN>' \\",
            f"  --data-raw {shlex.quote(payload_raw)}",
        ]
    )


def _normalize_product_name(value: str) -> str:
    normalized = value.strip().lower()
    canonical = _PRODUCT_NAME_BY_ALIAS.get(normalized)
    if canonical is None:
        accepted = ", ".join(sorted(_SCOPE_CATALOG))
        raise typer.BadParameter(f"Unknown product {value!r}. Expected one of: {accepted}")
    return canonical


def _select_products(*, preset_products: list[str] | None, interactive: bool) -> list[str]:
    if preset_products:
        return _dedupe([_normalize_product_name(item) for item in preset_products])

    if not interactive:
        raise typer.BadParameter("At least one --product is required when using --no-interactive")

    selected: list[str] = []
    for product in _SCOPE_CATALOG:
        if typer.confirm(f"Include {product} scopes?", default=(product == "CRM")):
            selected.append(product)

    if not selected:
        raise typer.BadParameter("At least one product must be selected")
    return selected


def _resolve_access_profile(*, access: AccessProfile | None, interactive: bool) -> AccessProfile:
    if access is not None:
        return access
    if not interactive:
        raise typer.BadParameter("--access is required when using --no-interactive")

    while True:
        selected = typer.prompt("Access profile [read/write/admin]", default="read").strip().lower()
        try:
            return AccessProfile(selected)
        except ValueError:
            typer.echo("Invalid access profile. Use read, write, or admin.", err=True)


def _resolve_include_common(*, include_common: bool | None, interactive: bool) -> bool:
    if include_common is not None:
        return include_common
    if interactive:
        return typer.confirm(
            "Include common scope examples for each selected product?", default=False
        )
    return False


def _resolve_additional_scopes(
    *,
    additional_scope: list[str] | None,
    interactive: bool,
) -> list[str]:
    extras = _validate_scope_tokens(_extract_scopes(scopes_csv=None, scope_items=additional_scope))
    if extras:
        return extras
    if not interactive:
        return []

    prompted = typer.prompt(
        "Additional scopes (comma-separated, optional)",
        default="",
        show_default=False,
    )
    return _validate_scope_tokens(_extract_scopes(scopes_csv=prompted, scope_items=None))


def _scopes_for_product(
    *,
    product: str,
    access: AccessProfile,
    include_common: bool,
) -> list[str]:
    spec = _SCOPE_CATALOG[product]
    if access is AccessProfile.read:
        scopes = list(spec.read)
    elif access is AccessProfile.write:
        scopes = list(spec.write)
    else:
        scopes = list(spec.write)
        scopes.extend([scope for scope in spec.common if ".ALL" in scope])

    if include_common:
        scopes.extend(spec.common)
    return _dedupe(scopes)


def _render_scope_output(*, payload: dict[str, Any], output_format: OutputFormat) -> str:
    scopes = payload["scopes"]
    scopes_csv = payload["scopes_csv"]

    if output_format is OutputFormat.json:
        return json.dumps(payload, indent=2, sort_keys=True)
    if output_format is OutputFormat.text:
        lines = [
            "Selected products: " + ", ".join(payload["products"]),
            f"Access profile: {payload['access_profile']}",
            f"Scope count: {payload['scope_count']}",
            "Scopes:",
        ]
        lines.extend(f"- {scope}" for scope in scopes)
        lines.extend(["", f"CSV: {scopes_csv}"])
        return "\n".join(lines)

    return "\n".join(
        [
            f'export ZOHO_SCOPES="{scopes_csv}"',
            'echo "$ZOHO_SCOPES"',
            'uv run zoho-auth grant-code --self-client-id <id> --scopes "$ZOHO_SCOPES"',
        ]
    )


@app.command("exchange-token")
def exchange_token(
    grant_code: Annotated[str, typer.Option("--grant-code", help="Authorization grant code.")],
    credentials_file: Annotated[
        str | None,
        typer.Option(
            "--credentials-file",
            help="Optional .env override path. Defaults to ZOHO_CREDENTIALS_FILE.",
        ),
    ] = None,
    accounts_domain: Annotated[
        str | None,
        typer.Option("--accounts-domain", help="Accounts domain override."),
    ] = None,
    redirect_uri: Annotated[
        str | None,
        typer.Option("--redirect-uri", help="Optional redirect URI."),
    ] = None,
    timeout: Annotated[float, typer.Option("--timeout", min=1.0)] = 30.0,
    show_secrets: Annotated[bool, typer.Option("--show-secrets")] = False,
) -> None:
    """Exchange an authorization code for OAuth tokens."""

    _load_credentials_env(credentials_file)
    client_id = _required_env("ZOHO_CLIENT_ID")
    client_secret = _required_env("ZOHO_CLIENT_SECRET")
    resolved_accounts_domain = _resolve_accounts_domain(accounts_domain)

    payload: dict[str, str] = {
        "grant_type": "authorization_code",
        "client_id": client_id,
        "client_secret": client_secret,
        "code": grant_code,
    }
    resolved_redirect_uri = redirect_uri or os.getenv("ZOHO_REDIRECT_URI")
    if resolved_redirect_uri:
        payload["redirect_uri"] = resolved_redirect_uri

    token_url = f"{resolved_accounts_domain}/oauth/v2/token"
    with httpx.Client(timeout=timeout) as http:
        response = http.post(
            token_url,
            data=payload,
            headers={"User-Agent": "zoho-auth-cli/0.1.3"},
        )

    body = _http_response_to_json(response)
    safe = _sanitize_payload(body, show_secrets=show_secrets)
    _print_json(safe)
    if response.status_code >= 400:
        raise typer.Exit(code=1)


@app.command("grant-code")
def grant_code(
    self_client_id: Annotated[
        str | None,
        typer.Option("--self-client-id", help="Self-client identifier."),
    ] = None,
    scopes: Annotated[
        str | None,
        typer.Option("--scopes", help="Comma-separated scopes."),
    ] = None,
    scope: Annotated[
        list[str] | None,
        typer.Option("--scope", help="Repeatable scope option."),
    ] = None,
    expiry: Annotated[int, typer.Option("--expiry", min=1)] = 10,
    description: Annotated[str, typer.Option("--description")] = "Evaluation of Zoho SDK",
    approved_orgs: Annotated[str | None, typer.Option("--approved-orgs")] = None,
    api_console_domain: Annotated[
        str,
        typer.Option("--api-console-domain", help="API Console domain."),
    ] = "https://api-console.zoho.com",
    execute: Annotated[bool, typer.Option("--execute")] = False,
    session_cookie: Annotated[
        str | None,
        typer.Option("--session-cookie", help="Session cookie for execute mode."),
    ] = None,
    x_zcsrf_token: Annotated[
        str | None,
        typer.Option("--x-zcsrf-token", help="x-zcsrf-token header."),
    ] = None,
    header: Annotated[
        list[str] | None,
        typer.Option("--header", help="Extra header in KEY=VALUE form."),
    ] = None,
    timeout: Annotated[float, typer.Option("--timeout", min=1.0)] = 30.0,
    show_secrets: Annotated[bool, typer.Option("--show-secrets")] = False,
    credentials_file: Annotated[
        str | None,
        typer.Option(
            "--credentials-file",
            help="Optional .env override path. Defaults to ZOHO_CREDENTIALS_FILE.",
        ),
    ] = None,
) -> None:
    """Build or execute a self-client granttoken request."""

    _load_credentials_env(credentials_file)
    resolved_client_id = self_client_id or os.getenv("ZOHO_SELF_CLIENT_ID")
    if not resolved_client_id:
        raise typer.BadParameter(
            "self-client id is required (--self-client-id or ZOHO_SELF_CLIENT_ID)"
        )

    payload = _build_grant_payload(
        scopes_csv=scopes,
        scope_items=scope,
        expiry=expiry,
        description=description,
        approved_orgs=approved_orgs,
    )
    endpoint = _grant_endpoint(api_console_domain, resolved_client_id)

    if not execute:
        _print_json(
            {
                "mode": "template",
                "endpoint": endpoint,
                "payload": payload,
                "curl_template": _render_grant_curl_template(endpoint, payload),
            }
        )
        return

    if not session_cookie:
        raise typer.BadParameter("--session-cookie is required when using --execute")

    headers = {
        "Accept": "*/*",
        "Content-Type": "application/json",
        "Cookie": session_cookie,
        "User-Agent": "zoho-auth-cli/0.1.3",
    }
    if x_zcsrf_token:
        headers["x-zcsrf-token"] = x_zcsrf_token
    headers.update(_parse_extra_headers(header))

    with httpx.Client(timeout=timeout) as http:
        response = http.post(endpoint, headers=headers, json=payload)

    body = _http_response_to_json(response)
    safe = _sanitize_payload(body, show_secrets=show_secrets)
    _print_json(safe)
    if response.status_code >= 400:
        raise typer.Exit(code=1)


@app.command("scope-builder")
def scope_builder(
    product: Annotated[
        list[str] | None,
        typer.Option(
            "--product",
            "-p",
            help="Repeatable product name: CRM, Creator, Projects, People, Sheet, WorkDrive.",
        ),
    ] = None,
    access: Annotated[
        AccessProfile | None,
        typer.Option("--access", help="Scope profile: read, write, or admin."),
    ] = None,
    include_common: Annotated[
        bool | None,
        typer.Option(
            "--include-common/--no-include-common",
            help="Include broader common scope examples for each selected product.",
        ),
    ] = None,
    scope: Annotated[
        list[str] | None,
        typer.Option("--scope", help="Repeatable additional scopes to include."),
    ] = None,
    interactive: Annotated[
        bool,
        typer.Option("--interactive/--no-interactive", help="Enable interactive prompts."),
    ] = True,
    output_format: Annotated[
        OutputFormat,
        typer.Option("--format", help="Output format: json, text, or env."),
    ] = OutputFormat.json,
    output_file: Annotated[
        Path | None,
        typer.Option("--output-file", help="Optional path to write output."),
    ] = None,
) -> None:
    """Interactive scope builder for selecting product scope bundles."""

    selected_products = _select_products(preset_products=product, interactive=interactive)
    selected_access = _resolve_access_profile(access=access, interactive=interactive)
    selected_include_common = _resolve_include_common(
        include_common=include_common,
        interactive=interactive,
    )
    extra_scopes = _resolve_additional_scopes(
        additional_scope=scope,
        interactive=interactive,
    )

    scopes_by_product: dict[str, list[str]] = {}
    combined_scopes: list[str] = []

    for selected_product in selected_products:
        product_scopes = _scopes_for_product(
            product=selected_product,
            access=selected_access,
            include_common=selected_include_common,
        )
        scopes_by_product[selected_product] = product_scopes
        combined_scopes.extend(product_scopes)

    combined_scopes.extend(extra_scopes)
    final_scopes = _validate_scope_tokens(_dedupe(combined_scopes))

    payload: dict[str, Any] = {
        "products": selected_products,
        "access_profile": selected_access.value,
        "include_common": selected_include_common,
        "scopes_by_product": scopes_by_product,
        "scope_count": len(final_scopes),
        "scopes": final_scopes,
        "scopes_csv": ",".join(final_scopes),
    }

    rendered = _render_scope_output(payload=payload, output_format=output_format)
    typer.echo(rendered)

    if output_file is not None:
        output_file.parent.mkdir(parents=True, exist_ok=True)
        output_file.write_text(rendered + "\n", encoding="utf-8")
        typer.echo(f"Wrote scope output to {output_file}")


def main(argv: list[str] | None = None) -> int:
    command_args = argv if argv is not None else sys.argv[1:]
    try:
        app(args=command_args, prog_name="zoho-auth", standalone_mode=False)
    except typer.Exit as exc:
        return int(exc.exit_code)
    except click.ClickException as exc:
        exc.show()
        return int(exc.exit_code)
    except Exception as exc:  # pragma: no cover - top-level safeguard
        typer.echo(f"error: {exc}", err=True)
        return 2
    return 0


if __name__ == "__main__":
    raise SystemExit(main(sys.argv[1:]))



"""Runtime CRM submodule discovery and dynamic module clients."""

from __future__ import annotations

import re
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, cast

from zoho.core.discovery_cache import DiscoveryDiskCache
from zoho.crm.models import ActionResponse, Record, RecordListResponse

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from zoho.crm.client import CRMClient


_INVALID_CLASS_CHAR_PATTERN = re.compile(r"[^0-9A-Za-z_]+")


class CRMDynamicModuleClient:
    """Runtime-bound CRM module client.

    This wrapper keeps the high-usage record methods, but bakes in a module API name.
    """

    def __init__(
        self,
        crm_client: CRMClient,
        *,
        module_api_name: str,
        module_metadata: Mapping[str, Any] | None = None,
    ) -> None:
        self._crm = crm_client
        self._module_api_name = module_api_name
        self._module_metadata = dict(module_metadata or {})

    @property
    def module_api_name(self) -> str:
        return self._module_api_name

    @property
    def module_metadata(self) -> dict[str, Any]:
        return dict(self._module_metadata)

    async def get(
        self,
        *,
        record_id: str | int,
        fields: Sequence[str] | None = None,
    ) -> Record:
        return await self._crm.records.get(
            module=self._module_api_name,
            record_id=record_id,
            fields=fields,
        )

    async def list(
        self,
        *,
        page: int = 1,
        per_page: int = 200,
        fields: Sequence[str] | None = None,
        extra_params: Mapping[str, Any] | None = None,
    ) -> RecordListResponse:
        return await self._crm.records.list(
            module=self._module_api_name,
            page=page,
            per_page=per_page,
            fields=fields,
            extra_params=extra_params,
        )

    async def iter(
        self,
        *,
        per_page: int = 200,
        fields: Sequence[str] | None = None,
        extra_params: Mapping[str, Any] | None = None,
    ) -> AsyncIterator[Record]:
        async for record in self._crm.records.iter(
            module=self._module_api_name,
            per_page=per_page,
            fields=fields,
            extra_params=extra_params,
        ):
            yield record

    async def create(
        self,
        *,
        data: Mapping[str, Any] | Sequence[Mapping[str, Any]],
        trigger: Sequence[str] | None = None,
    ) -> ActionResponse:
        return await self._crm.records.create(
            module=self._module_api_name,
            data=data,
            trigger=trigger,
        )

    async def update(
        self,
        *,
        record_id: str | int,
        data: Mapping[str, Any],
        trigger: Sequence[str] | None = None,
    ) -> ActionResponse:
        return await self._crm.records.update(
            module=self._module_api_name,
            record_id=record_id,
            data=data,
            trigger=trigger,
        )

    async def delete(self, *, record_id: str | int) -> ActionResponse:
        return await self._crm.records.delete(module=self._module_api_name, record_id=record_id)


def _dynamic_module_class_name(module_name: str) -> str:
    clean = _INVALID_CLASS_CHAR_PATTERN.sub("_", module_name).strip("_")
    if not clean:
        clean = "Module"
    if clean[0].isdigit():
        clean = f"_{clean}"
    return f"CRM{clean}ModuleClient"


class CRMDynamicNamespace:
    """Dynamic CRM namespace with runtime module discovery."""

    def __init__(
        self,
        crm_client: CRMClient,
        *,
        discovery_cache: DiscoveryDiskCache | None = None,
        discovery_cache_scope: str = "default:US:production",
    ) -> None:
        self._crm = crm_client
        self._clients: dict[str, CRMDynamicModuleClient] = {}
        self._dynamic_classes: dict[str, type[CRMDynamicModuleClient]] = {}
        self._module_metadata_by_name: dict[str, dict[str, Any]] = {}
        self._module_name_aliases: dict[str, str] = {}
        self._module_names: list[str] = []
        self._discovery_cache = discovery_cache
        self._discovery_cache_scope = discovery_cache_scope

    async def list_modules(
        self,
        *,
        use_cache: bool = True,
        cache_ttl_seconds: int | None = None,
    ) -> list[str]:
        if use_cache and self._module_names:
            return list(self._module_names)

        if use_cache and self._discovery_cache is not None:
            cached_modules = self._discovery_cache.load(
                product="crm",
                resource="modules",
                scope=self._discovery_cache_scope,
            )
            if cached_modules:
                self._ingest_module_metadata(cached_modules)
                return list(self._module_names)

        modules = await self._crm.modules.list(
            use_cache=use_cache,
            cache_ttl_seconds=cache_ttl_seconds,
        )
        self._ingest_module_metadata(modules)

        if self._discovery_cache is not None and modules:
            self._discovery_cache.save(
                product="crm",
                resource="modules",
                scope=self._discovery_cache_scope,
                entries=modules,
            )

        return list(self._module_names)

    async def has_module(
        self,
        name: str,
        *,
        use_cache: bool = True,
        cache_ttl_seconds: int | None = None,
    ) -> bool:
        await self.list_modules(use_cache=use_cache, cache_ttl_seconds=cache_ttl_seconds)
        return name.lower() in self._module_name_aliases

    async def get_module_client(
        self,
        name: str,
        *,
        use_cache: bool = True,
        cache_ttl_seconds: int | None = None,
    ) -> CRMDynamicModuleClient:
        await self.list_modules(use_cache=use_cache, cache_ttl_seconds=cache_ttl_seconds)
        canonical_name = self._module_name_aliases.get(name.lower())
        if canonical_name is None:
            sample = ", ".join(sorted(self._module_metadata_by_name)[:8])
            message = f"Unknown CRM module: {name!r}"
            if sample:
                message = f"{message}. Known modules include: {sample}"
            raise KeyError(message)
        return self._client_for_module(canonical_name)

    async def precompile_modules(
        self,
        *,
        cache_ttl_seconds: int | None = None,
    ) -> list[str]:
        """Force-refresh CRM module metadata and write it to disk cache."""

        return await self.list_modules(use_cache=False, cache_ttl_seconds=cache_ttl_seconds)

    def _client_for_module(self, module_name: str) -> CRMDynamicModuleClient:
        existing = self._clients.get(module_name)
        if existing is not None:
            return existing

        dynamic_class = self._dynamic_classes.get(module_name)
        if dynamic_class is None:
            class_name = _dynamic_module_class_name(module_name)
            dynamic_class = cast(
                type[CRMDynamicModuleClient],
                type(
                    class_name,
                    (CRMDynamicModuleClient,),
                    {
                        "__doc__": (
                            f"Runtime CRM module client for {module_name!r}. "
                            "Methods mirror `client.crm.records`."
                        )
                    },
                ),
            )
            self._dynamic_classes[module_name] = dynamic_class

        client = dynamic_class(
            self._crm,
            module_api_name=module_name,
            module_metadata=self._module_metadata_by_name.get(module_name),
        )
        self._clients[module_name] = client
        return client

    def __getattr__(self, name: str) -> CRMDynamicModuleClient:
        if name.startswith("_"):
            raise AttributeError(f"{self.__class__.__name__} has no attribute {name!r}")
        canonical_name = self._module_name_aliases.get(name.lower(), name)
        return self._client_for_module(canonical_name)

    def _ingest_module_metadata(self, modules: Sequence[Mapping[str, Any]]) -> None:
        names: list[str] = []
        metadata_map: dict[str, dict[str, Any]] = {}
        aliases: dict[str, str] = {}

        for module in modules:
            name_raw = module.get("api_name")
            if not isinstance(name_raw, str) or not name_raw:
                continue
            names.append(name_raw)
            metadata_map[name_raw] = dict(module)
            aliases[name_raw.lower()] = name_raw

        self._module_names = names
        self._module_metadata_by_name = metadata_map
        self._module_name_aliases = aliases



"""Top-level Zoho SDK client."""

from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from zoho.connections import ZohoConnectionProfile, ZohoConnectionsManager
from zoho.core.auth import OAuth2RefreshAuthProvider
from zoho.core.cache import AsyncTTLCache
from zoho.core.discovery_cache import DiscoveryDiskCache
from zoho.core.logging import configure_logging, get_logger
from zoho.core.token_store import TokenStore, build_token_store
from zoho.core.transport import HttpxTransport
from zoho.settings import (
    CacheSettings,
    CreatorEnvironmentHeader,
    DataCenter,
    EnvironmentName,
    LogFormat,
    RetrySettings,
    TokenStoreBackend,
    TransportSettings,
    ZohoSettings,
    _default_discovery_cache_dir,
    _default_token_store_path,
)

if TYPE_CHECKING:
    from zoho.analytics.client import AnalyticsClient
    from zoho.cliq.client import CliqClient
    from zoho.creator.client import CreatorClient
    from zoho.crm.client import CRMClient
    from zoho.mail.client import MailClient
    from zoho.people.client import PeopleClient
    from zoho.projects.client import ProjectsClient
    from zoho.sheet.client import SheetClient
    from zoho.workdrive.client import WorkDriveClient
    from zoho.writer.client import WriterClient

_PROJECTS_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://projectsapi.zoho.com",
    "EU": "https://projectsapi.zoho.eu",
    "IN": "https://projectsapi.zoho.in",
    "AU": "https://projectsapi.zoho.com.au",
    "JP": "https://projectsapi.zoho.jp",
    "CA": "https://projectsapi.zohocloud.ca",
    "SA": "https://projectsapi.zoho.sa",
    "CN": "https://projectsapi.zoho.com.cn",
}

_PEOPLE_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://people.zoho.com",
    "EU": "https://people.zoho.eu",
    "IN": "https://people.zoho.in",
    "AU": "https://people.zoho.com.au",
    "JP": "https://people.zoho.jp",
    "CA": "https://people.zohocloud.ca",
    "SA": "https://people.zoho.sa",
    "CN": "https://people.zoho.com.cn",
}

_SHEET_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://sheet.zoho.com",
    "EU": "https://sheet.zoho.eu",
    "IN": "https://sheet.zoho.in",
    "AU": "https://sheet.zoho.com.au",
    "JP": "https://sheet.zoho.jp",
    "CA": "https://sheet.zohocloud.ca",
    "SA": "https://sheet.zoho.sa",
    "CN": "https://sheet.zoho.com.cn",
}

_WORKDRIVE_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://www.zohoapis.com",
    "EU": "https://www.zohoapis.eu",
    "IN": "https://www.zohoapis.in",
    "AU": "https://www.zohoapis.com.au",
    "JP": "https://www.zohoapis.jp",
    "CA": "https://www.zohoapis.ca",
    "SA": "https://www.zohoapis.sa",
    "CN": "https://www.zohoapis.com.cn",
}

_CLIQ_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://cliq.zoho.com",
    "EU": "https://cliq.zoho.eu",
    "IN": "https://cliq.zoho.in",
    "AU": "https://cliq.zoho.com.au",
    "JP": "https://cliq.zoho.jp",
    "CA": "https://cliq.zohocloud.ca",
    "SA": "https://cliq.zoho.sa",
    "CN": "https://cliq.zoho.com.cn",
}

_ANALYTICS_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://analyticsapi.zoho.com",
    "EU": "https://analyticsapi.zoho.eu",
    "IN": "https://analyticsapi.zoho.in",
    "AU": "https://analyticsapi.zoho.com.au",
    "JP": "https://analyticsapi.zoho.jp",
    "CA": "https://analyticsapi.zohocloud.ca",
    "SA": "https://analyticsapi.zoho.sa",
    "CN": "https://analyticsapi.zoho.com.cn",
}

_WRITER_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://www.zohoapis.com",
    "EU": "https://www.zohoapis.eu",
    "IN": "https://www.zohoapis.in",
    "AU": "https://www.zohoapis.com.au",
    "JP": "https://www.zohoapis.jp",
    "CA": "https://www.zohoapis.ca",
    "SA": "https://www.zohoapis.sa",
    "CN": "https://www.zohoapis.com.cn",
}

_MAIL_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://mail.zoho.com",
    "EU": "https://mail.zoho.eu",
    "IN": "https://mail.zoho.in",
    "AU": "https://mail.zoho.com.au",
    "JP": "https://mail.zoho.jp",
    "CA": "https://mail.zohocloud.ca",
    "SA": "https://mail.zoho.sa",
    "CN": "https://mail.zoho.com.cn",
}

_DEFAULT_TOKEN_STORE_PATH = _default_token_store_path()
_DEFAULT_DISCOVERY_CACHE_DIR = _default_discovery_cache_dir()


class Zoho:
    """Async first SDK entry point.

    Example:
        ```python
        from zoho import Zoho

        # one-shot scripts / jobs
        async with Zoho.from_credentials(
            client_id="...",
            client_secret="...",
            refresh_token="...",
        ) as zoho:
            lead = await zoho.crm.records.get(module="Leads", record_id="123456789")
            print(lead.id)

        # long-lived app singleton
        zoho = Zoho.from_credentials(
            client_id="...",
            client_secret="...",
            refresh_token="...",
        )
        lead = await zoho.crm.records.get(module="Leads", record_id="123456789")
        await zoho.close()
        ```
    """

    def __init__(self, settings: ZohoSettings, *, token_store: TokenStore | None = None) -> None:
        self.settings = settings

        configure_logging(log_format=settings.log_format, log_level=settings.log_level)
        self._logger = get_logger("zoho.client")

        self._token_store = token_store or build_token_store(
            backend=settings.token_store_backend,
            sqlite_path=settings.token_store_path,
            redis_url=settings.redis_url,
        )

        self._auth = OAuth2RefreshAuthProvider(
            client_id=settings.client_id,
            client_secret=settings.client_secret,
            refresh_token=settings.refresh_token,
            dc=settings.dc,
            environment=settings.environment,
            token_store=self._token_store,
            accounts_domain=settings.accounts_domain,
            api_domain=settings.api_domain,
            cache_namespace=settings.connection_name,
            user_agent=settings.user_agent,
        )

        self._transport = HttpxTransport(
            timeout_seconds=settings.transport.timeout_seconds,
            max_connections=settings.transport.max_connections,
            max_keepalive_connections=settings.transport.max_keepalive_connections,
            verify_ssl=settings.transport.verify_ssl,
            max_retries=settings.retry.max_retries,
            backoff_base_seconds=settings.retry.backoff_base_seconds,
            backoff_max_seconds=settings.retry.backoff_max_seconds,
            retry_status_codes=settings.retry.retry_status_codes,
        )

        self._metadata_cache: AsyncTTLCache[dict[str, Any]] | None = None
        if settings.cache.enable_metadata_cache:
            self._metadata_cache = AsyncTTLCache()
        self._discovery_cache: DiscoveryDiskCache | None = None
        if settings.cache.enable_discovery_cache:
            self._discovery_cache = DiscoveryDiskCache(
                base_dir=settings.cache.discovery_cache_dir,
                ttl_seconds=settings.cache.discovery_cache_ttl_seconds,
            )
        self._discovery_cache_scope = (
            f"{settings.connection_name}:{settings.dc}:{settings.environment}"
        )

        self._crm: CRMClient | None = None
        self._creator: CreatorClient | None = None
        self._projects: ProjectsClient | None = None
        self._people: PeopleClient | None = None
        self._sheet: SheetClient | None = None
        self._workdrive: WorkDriveClient | None = None
        self._cliq: CliqClient | None = None
        self._analytics: AnalyticsClient | None = None
        self._writer: WriterClient | None = None
        self._mail: MailClient | None = None

        self._connections: ZohoConnectionsManager | None = None
        self._closed = False

    @classmethod
    def from_settings(cls, settings: ZohoSettings) -> Zoho:
        return cls(settings)

    @classmethod
    def from_env(cls) -> Zoho:
        """Create a client from `ZOHO_*` environment variables."""

        return cls(ZohoSettings())  # type: ignore[call-arg]

    @classmethod
    def from_credentials(
        cls,
        *,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        dc: DataCenter = "US",
        environment: EnvironmentName = "production",
        connection_name: str = "default",
        accounts_domain: str | None = None,
        api_domain: str | None = None,
        creator_base_url: str | None = None,
        creator_environment_header: CreatorEnvironmentHeader | None = None,
        projects_base_url: str | None = None,
        projects_default_portal_id: str | None = None,
        people_base_url: str | None = None,
        sheet_base_url: str | None = None,
        workdrive_base_url: str | None = None,
        cliq_base_url: str | None = None,
        analytics_base_url: str | None = None,
        writer_base_url: str | None = None,
        mail_base_url: str | None = None,
        token_store_backend: TokenStoreBackend = "sqlite",
        token_store_path: Path = _DEFAULT_TOKEN_STORE_PATH,
        redis_url: str | None = None,
        log_format: LogFormat = "pretty",
        log_level: str = "INFO",
        user_agent: str = "zoho-sdk/0.1.3",
        timeout_seconds: float = 30.0,
        max_connections: int = 100,
        max_keepalive_connections: int = 20,
        verify_ssl: bool = True,
        max_retries: int = 3,
        backoff_base_seconds: float = 0.25,
        backoff_max_seconds: float = 5.0,
        enable_metadata_cache: bool = True,
        metadata_cache_ttl_seconds: int = 24 * 60 * 60,
        enable_discovery_cache: bool = True,
        discovery_cache_ttl_seconds: int = 24 * 60 * 60,
        discovery_cache_dir: Path = _DEFAULT_DISCOVERY_CACHE_DIR,
    ) -> Zoho:
        """Create a client with explicit configuration.

        This constructor is the recommended onboarding path for production usage.
        """

        settings = ZohoSettings(
            connection_name=connection_name,
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
            dc=dc,
            environment=environment,
            accounts_domain=accounts_domain,
            api_domain=api_domain,
            creator_base_url=creator_base_url,
            creator_environment_header=creator_environment_header,
            projects_base_url=projects_base_url,
            projects_default_portal_id=projects_default_portal_id,
            people_base_url=people_base_url,
            sheet_base_url=sheet_base_url,
            workdrive_base_url=workdrive_base_url,
            cliq_base_url=cliq_base_url,
            analytics_base_url=analytics_base_url,
            writer_base_url=writer_base_url,
            mail_base_url=mail_base_url,
            token_store_backend=token_store_backend,
            token_store_path=token_store_path.expanduser(),
            redis_url=redis_url,
            log_format=log_format,
            log_level=log_level,
            user_agent=user_agent,
            transport=TransportSettings(
                timeout_seconds=timeout_seconds,
                max_connections=max_connections,
                max_keepalive_connections=max_keepalive_connections,
                verify_ssl=verify_ssl,
            ),
            retry=RetrySettings(
                max_retries=max_retries,
                backoff_base_seconds=backoff_base_seconds,
                backoff_max_seconds=backoff_max_seconds,
            ),
            cache=CacheSettings(
                enable_metadata_cache=enable_metadata_cache,
                metadata_cache_ttl_seconds=metadata_cache_ttl_seconds,
                enable_discovery_cache=enable_discovery_cache,
                discovery_cache_ttl_seconds=discovery_cache_ttl_seconds,
                discovery_cache_dir=discovery_cache_dir.expanduser(),
            ),
        )
        return cls(settings)

    @property
    def crm(self) -> CRMClient:
        """Access Zoho CRM operations."""

        if self._crm is None:
            from zoho.crm.client import CRMClient

            self._crm = CRMClient(
                request=self._request_crm,
                metadata_cache=self._metadata_cache,
                default_metadata_ttl_seconds=self.settings.cache.metadata_cache_ttl_seconds,
                discovery_cache=self._discovery_cache,
                discovery_cache_scope=self._discovery_cache_scope,
            )
        return self._crm

    @property
    def creator(self) -> CreatorClient:
        """Access Zoho Creator operations."""

        if self._creator is None:
            from zoho.creator.client import CreatorClient

            self._creator = CreatorClient(
                request=self._request_creator,
                discovery_cache=self._discovery_cache,
                discovery_cache_scope=self._discovery_cache_scope,
            )
        return self._creator

    @property
    def projects(self) -> ProjectsClient:
        """Access Zoho Projects operations."""

        if self._projects is None:
            from zoho.projects.client import ProjectsClient

            self._projects = ProjectsClient(
                request=self._request_projects,
                default_portal_id=self.settings.projects_default_portal_id,
            )
        return self._projects

    @property
    def people(self) -> PeopleClient:
        """Access Zoho People operations."""

        if self._people is None:
            from zoho.people.client import PeopleClient

            self._people = PeopleClient(request=self._request_people)
        return self._people

    @property
    def sheet(self) -> SheetClient:
        """Access Zoho Sheet operations."""

        if self._sheet is None:
            from zoho.sheet.client import SheetClient

            self._sheet = SheetClient(request=self._request_sheet)
        return self._sheet

    @property
    def workdrive(self) -> WorkDriveClient:
        """Access Zoho WorkDrive operations."""

        if self._workdrive is None:
            from zoho.workdrive.client import WorkDriveClient

            self._workdrive = WorkDriveClient(request=self._request_workdrive)
        return self._workdrive

    @property
    def cliq(self) -> CliqClient:
        """Access Zoho Cliq operations."""

        if self._cliq is None:
            from zoho.cliq.client import CliqClient

            self._cliq = CliqClient(request=self._request_cliq)
        return self._cliq

    @property
    def analytics(self) -> AnalyticsClient:
        """Access Zoho Analytics operations."""

        if self._analytics is None:
            from zoho.analytics.client import AnalyticsClient

            self._analytics = AnalyticsClient(request=self._request_analytics)
        return self._analytics

    @property
    def writer(self) -> WriterClient:
        """Access Zoho Writer operations."""

        if self._writer is None:
            from zoho.writer.client import WriterClient

            self._writer = WriterClient(request=self._request_writer)
        return self._writer

    @property
    def mail(self) -> MailClient:
        """Access Zoho Mail operations."""

        if self._mail is None:
            from zoho.mail.client import MailClient

            self._mail = MailClient(request=self._request_mail)
        return self._mail

    @property
    def connections(self) -> ZohoConnectionsManager:
        """Access named connection profiles for multi-account usage."""

        if self._connections is None:
            self._connections = ZohoConnectionsManager(self)
        return self._connections

    def for_connection(self, name: str) -> Zoho:
        """Get a client view bound to a named connection profile."""

        if name == "default":
            return self
        return self.connections.get(name)

    def register_connection(self, profile: ZohoConnectionProfile) -> None:
        """Register an additional connection profile on this client."""

        self.connections.register(profile)

    @property
    def closed(self) -> bool:
        """Whether the client lifecycle has been closed."""

        return self._closed

    def _ensure_open(self) -> None:
        if self._closed:
            raise RuntimeError(
                "Zoho client is closed. Create a new client instance or keep a long-lived "
                "singleton open until app shutdown."
            )

    async def _request_crm(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        api_domain = await self._auth.get_api_domain()
        url = f"{api_domain}/crm/v8{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_creator(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (self.settings.creator_base_url or await self._auth.get_api_domain()).rstrip(
            "/"
        )
        if clean_path.startswith("/creator/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/creator/v2.1{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers: dict[str, str] = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
        }
        if self.settings.creator_environment_header is not None:
            request_headers["environment"] = self.settings.creator_environment_header
        request_headers.update(dict(headers or {}))

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_projects(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.projects_base_url or _PROJECTS_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/api/v3/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/api/v3{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_people(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.people_base_url or _PEOPLE_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/people/") or clean_path.startswith("/api/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/people/api{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_sheet(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.sheet_base_url or _SHEET_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/api/v2/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/api/v2{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_workdrive(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.workdrive_base_url or _WORKDRIVE_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/workdrive/"):
            url = f"{base_domain}{clean_path}"
        elif clean_path.startswith("/api/v1/"):
            url = f"{base_domain}/workdrive{clean_path}"
        else:
            url = f"{base_domain}/workdrive/api/v1{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_cliq(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.cliq_base_url or _CLIQ_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/api/v2/") or clean_path.startswith("/network/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/api/v2{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_analytics(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.analytics_base_url or _ANALYTICS_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/restapi/v2/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/restapi/v2{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_writer(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.writer_base_url or _WRITER_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/writer/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/writer/api/v1{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    async def _request_mail(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        self._ensure_open()
        clean_path = path if path.startswith("/") else f"/{path}"

        base_domain = (
            self.settings.mail_base_url or _MAIL_API_DOMAIN_BY_DC[self.settings.dc]
        ).rstrip("/")
        if clean_path.startswith("/api/"):
            url = f"{base_domain}{clean_path}"
        else:
            url = f"{base_domain}/api{clean_path}"

        auth_headers = await self._auth.get_auth_headers()
        request_headers = {
            "Accept": "application/json",
            "User-Agent": self.settings.user_agent,
            **auth_headers,
            **dict(headers or {}),
        }

        response = await self._transport.request(
            method,
            url,
            headers=request_headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )
        return self._parse_response(response)

    @staticmethod
    def _parse_response(response: Any) -> dict[str, Any]:
        if response.status_code == 204 or not response.content:
            return {}

        try:
            payload = response.json()
        except ValueError:
            return {"raw": response.text}

        if isinstance(payload, Mapping):
            return dict(payload)
        return {"data": cast(list[Any], payload)}

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True

        if self._connections is not None:
            await self._connections.close_all()

        await self._auth.close()
        await self._transport.close()
        self._logger.info("zoho_client_closed")

    async def aclose(self) -> None:
        await self.close()

    async def __aenter__(self) -> Zoho:
        self._ensure_open()
        return self

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        await self.close()



"""Sheet worksheet APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.sheet.models import SheetResponse, parse_sheet_response

if TYPE_CHECKING:
    from zoho.sheet.client import SheetClient


class SheetWorksheetsClient:
    """Worksheet operations for Zoho Sheet."""

    def __init__(self, sheet_client: SheetClient) -> None:
        self._sheet = sheet_client

    async def list(
        self,
        *,
        workbook_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request(
            "GET",
            f"/workbooks/{workbook_id}/worksheets",
            headers=headers,
        )
        return parse_sheet_response(payload)

    async def create(
        self,
        *,
        workbook_id: str,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request(
            "POST",
            f"/workbooks/{workbook_id}/worksheets",
            json=dict(data),
            headers=headers,
        )
        return parse_sheet_response(payload)

    async def rename(
        self,
        *,
        workbook_id: str,
        worksheet_id: str,
        name: str,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request(
            "PATCH",
            f"/workbooks/{workbook_id}/worksheets/{worksheet_id}",
            json={"name": name},
            headers=headers,
        )
        return parse_sheet_response(payload)

    async def delete(
        self,
        *,
        workbook_id: str,
        worksheet_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request(
            "DELETE",
            f"/workbooks/{workbook_id}/worksheets/{worksheet_id}",
            headers=headers,
        )
        return parse_sheet_response(payload)



"""CRM Record operations."""

from __future__ import annotations

from collections.abc import AsyncIterator, Mapping, Sequence
from typing import TYPE_CHECKING, Any

from zoho.core.pagination import AsyncPager, Page
from zoho.crm.models import ActionResponse, PageInfo, Record, RecordListResponse, extract_first_data

if TYPE_CHECKING:
    from zoho.crm.client import CRMClient


class RecordsClient:
    """High-usage CRM record operations.

    Example:
        ```python
        lead = await zoho.crm.records.get(module="Leads", record_id="123")
        await zoho.crm.records.update(module="Leads", record_id="123", data={"Last_Name": "Ng"})
        async for rec in zoho.crm.records.iter(module="Leads", fields=["Email", "Last_Name"]):
            print(rec["Email"])
        ```
    """

    def __init__(self, crm_client: CRMClient) -> None:
        self._crm = crm_client

    async def get(
        self,
        *,
        module: str,
        record_id: str | int,
        fields: Sequence[str] | None = None,
    ) -> Record:
        params: dict[str, Any] = {}
        if fields:
            params["fields"] = ",".join(fields)

        payload = await self._crm.request("GET", f"/{module}/{record_id}", params=params)
        return Record(extract_first_data(payload))

    async def list(
        self,
        *,
        module: str,
        page: int = 1,
        per_page: int = 200,
        fields: Sequence[str] | None = None,
        extra_params: Mapping[str, Any] | None = None,
    ) -> RecordListResponse:
        params: dict[str, Any] = {
            "page": page,
            "per_page": per_page,
        }
        if fields:
            params["fields"] = ",".join(fields)
        if extra_params:
            params.update(extra_params)

        payload = await self._crm.request("GET", f"/{module}", params=params)
        data_raw = payload.get("data")
        data = (
            [item for item in data_raw if isinstance(item, dict)]
            if isinstance(data_raw, list)
            else []
        )

        info_raw = payload.get("info")
        info = PageInfo.model_validate(info_raw) if isinstance(info_raw, dict) else None
        return RecordListResponse(data=data, info=info)

    async def iter(
        self,
        *,
        module: str,
        per_page: int = 200,
        fields: Sequence[str] | None = None,
        extra_params: Mapping[str, Any] | None = None,
    ) -> AsyncIterator[Record]:
        async def fetch_page(page: int, page_size: int) -> Page[Record]:
            result = await self.list(
                module=module,
                page=page,
                per_page=page_size,
                fields=fields,
                extra_params=extra_params,
            )
            records = [Record(item) for item in result.data]
            has_more = (
                bool(result.info.more_records)
                if result.info is not None
                else len(records) == page_size
            )
            next_page = page + 1 if has_more else None
            return Page(items=records, has_more=has_more, next_page=next_page)

        pager = AsyncPager(fetch_page=fetch_page, page_size=per_page)
        async for item in pager:
            yield item

    async def create(
        self,
        *,
        module: str,
        data: Mapping[str, Any] | Sequence[Mapping[str, Any]],
        trigger: Sequence[str] | None = None,
    ) -> ActionResponse:
        rows = [dict(data)] if isinstance(data, Mapping) else [dict(item) for item in data]
        payload: dict[str, Any] = {"data": rows}
        if trigger:
            payload["trigger"] = list(trigger)

        response = await self._crm.request("POST", f"/{module}", json=payload)
        return ActionResponse.model_validate(extract_first_data(response))

    async def update(
        self,
        *,
        module: str,
        record_id: str | int,
        data: Mapping[str, Any],
        trigger: Sequence[str] | None = None,
    ) -> ActionResponse:
        row = {"id": str(record_id), **dict(data)}
        payload: dict[str, Any] = {"data": [row]}
        if trigger:
            payload["trigger"] = list(trigger)

        response = await self._crm.request("PUT", f"/{module}", json=payload)
        return ActionResponse.model_validate(extract_first_data(response))

    async def delete(self, *, module: str, record_id: str | int) -> ActionResponse:
        response = await self._crm.request("DELETE", f"/{module}/{record_id}")
        return ActionResponse.model_validate(extract_first_data(response))



"""Zoho SDK public package.

This package is intentionally lazy-loaded for fast startup.
"""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.client import Zoho
    from zoho.connections import ZohoConnectionProfile
    from zoho.settings import ZohoSettings

__all__ = ["Zoho", "ZohoConnectionProfile", "ZohoSettings", "__version__"]

try:
    __version__ = version("zoho")
except PackageNotFoundError:  # pragma: no cover - local editable installs before metadata exists
    __version__ = "0.1.3"


def __getattr__(name: str) -> Any:
    if name == "Zoho":
        from zoho.client import Zoho

        return Zoho
    if name == "ZohoSettings":
        from zoho.settings import ZohoSettings

        return ZohoSettings
    if name == "ZohoConnectionProfile":
        from zoho.connections import ZohoConnectionProfile

        return ZohoConnectionProfile
    raise AttributeError(f"module 'zoho' has no attribute {name!r}")



"""CRM Modules metadata operations."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.crm.client import CRMClient


class ModulesClient:
    """CRM module metadata operations."""

    def __init__(self, crm_client: CRMClient) -> None:
        self._crm = crm_client

    async def list(
        self,
        *,
        use_cache: bool = True,
        cache_ttl_seconds: int | None = None,
    ) -> list[dict[str, Any]]:
        payload = await self._crm.request(
            "GET",
            "/settings/modules",
            use_cache=use_cache,
            cache_key="crm:settings:modules:list",
            cache_ttl_seconds=cache_ttl_seconds,
        )
        data = payload.get("modules")
        return [item for item in data if isinstance(item, dict)] if isinstance(data, list) else []

    async def get(
        self,
        module: str,
        *,
        use_cache: bool = True,
        cache_ttl_seconds: int | None = None,
    ) -> dict[str, Any]:
        payload = await self._crm.request(
            "GET",
            f"/settings/modules/{module}",
            use_cache=use_cache,
            cache_key=f"crm:settings:modules:get:{module}",
            cache_ttl_seconds=cache_ttl_seconds,
        )
        modules = payload.get("modules")
        if isinstance(modules, list) and modules and isinstance(modules[0], dict):
            return modules[0]
        return {}



"""Zoho People product client."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    from zoho.people.employees import PeopleEmployeesClient
    from zoho.people.files import PeopleFilesClient
    from zoho.people.forms import PeopleFormsClient


class PeopleRequestCallable(Protocol):
    async def __call__(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]: ...


class PeopleClient:
    """Product-scoped People client."""

    def __init__(self, *, request: PeopleRequestCallable) -> None:
        self._request_fn = request
        self._forms: PeopleFormsClient | None = None
        self._employees: PeopleEmployeesClient | None = None
        self._files: PeopleFilesClient | None = None

    @property
    def forms(self) -> PeopleFormsClient:
        if self._forms is None:
            from zoho.people.forms import PeopleFormsClient

            self._forms = PeopleFormsClient(self)
        return self._forms

    @property
    def employees(self) -> PeopleEmployeesClient:
        if self._employees is None:
            from zoho.people.employees import PeopleEmployeesClient

            self._employees = PeopleEmployeesClient(self)
        return self._employees

    @property
    def files(self) -> PeopleFilesClient:
        if self._files is None:
            from zoho.people.files import PeopleFilesClient

            self._files = PeopleFilesClient(self)
        return self._files

    async def request(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        return await self._request_fn(
            method,
            path,
            headers=headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )



"""Writer document APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.writer.models import WriterResponse, parse_writer_response

if TYPE_CHECKING:
    from zoho.writer.client import WriterClient


class WriterDocumentsClient:
    """Document operations for Zoho Writer."""

    def __init__(self, writer_client: WriterClient) -> None:
        self._writer = writer_client

    async def list(
        self,
        *,
        folder_id: str | None = None,
        page: int | None = None,
        limit: int | None = None,
        search: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        params: dict[str, Any] = {}
        if folder_id:
            params["folder_id"] = folder_id
        if page is not None:
            params["page"] = page
        if limit is not None:
            params["limit"] = limit
        if search:
            params["search"] = search

        payload = await self._writer.request("GET", "/documents", params=params, headers=headers)
        return parse_writer_response(payload)

    async def get(
        self,
        *,
        document_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        payload = await self._writer.request("GET", f"/documents/{document_id}", headers=headers)
        return parse_writer_response(payload)

    async def create(
        self,
        *,
        title: str,
        content: str | None = None,
        folder_id: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        data: dict[str, Any] = {"title": title}
        if content is not None:
            data["content"] = content
        if folder_id is not None:
            data["folder_id"] = folder_id

        payload = await self._writer.request("POST", "/documents", json=data, headers=headers)
        return parse_writer_response(payload)

    async def update(
        self,
        *,
        document_id: str,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        payload = await self._writer.request(
            "PUT",
            f"/documents/{document_id}",
            json=dict(data),
            headers=headers,
        )
        return parse_writer_response(payload)

    async def delete(
        self,
        *,
        document_id: str,
        permanent: bool = False,
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        payload = await self._writer.request(
            "DELETE",
            f"/documents/{document_id}",
            params={"permanent": permanent},
            headers=headers,
        )
        return parse_writer_response(payload)

    async def restore(
        self,
        *,
        document_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        payload = await self._writer.request(
            "POST",
            f"/documents/{document_id}/restore",
            headers=headers,
        )
        return parse_writer_response(payload)

    async def download(
        self,
        *,
        document_id: str,
        output_format: str = "docx",
        headers: Mapping[str, str] | None = None,
    ) -> WriterResponse:
        payload = await self._writer.request(
            "GET",
            f"/documents/{document_id}/export",
            params={"format": output_format},
            headers=headers,
        )
        return parse_writer_response(payload)



"""Sheet workbooks APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.sheet.models import SheetResponse, parse_sheet_response

if TYPE_CHECKING:
    from zoho.sheet.client import SheetClient


class SheetWorkbooksClient:
    """Workbook operations for Zoho Sheet."""

    def __init__(self, sheet_client: SheetClient) -> None:
        self._sheet = sheet_client

    async def list(
        self,
        *,
        from_index: int | None = None,
        limit: int | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        params: dict[str, Any] = {}
        if from_index is not None:
            params["from"] = from_index
        if limit is not None:
            params["limit"] = limit

        payload = await self._sheet.request("GET", "/workbooks", params=params, headers=headers)
        return parse_sheet_response(payload)

    async def get(
        self,
        *,
        workbook_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request("GET", f"/workbooks/{workbook_id}", headers=headers)
        return parse_sheet_response(payload)

    async def create(
        self,
        *,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request("POST", "/workbooks", json=dict(data), headers=headers)
        return parse_sheet_response(payload)

    async def delete(
        self,
        *,
        workbook_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> SheetResponse:
        payload = await self._sheet.request("DELETE", f"/workbooks/{workbook_id}", headers=headers)
        return parse_sheet_response(payload)



"""WorkDrive recent changes APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.workdrive.models import WorkDriveResponse, parse_workdrive_response

if TYPE_CHECKING:
    from zoho.workdrive.client import WorkDriveClient


class WorkDriveChangesClient:
    """Delta and recent-change operations for ingestion connectors."""

    def __init__(self, workdrive_client: WorkDriveClient) -> None:
        self._workdrive = workdrive_client

    async def list_recent(
        self,
        *,
        folder_id: str,
        cursor: str | None = None,
        limit: int | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> WorkDriveResponse:
        params: dict[str, Any] = {}
        if cursor is not None:
            params["page[next]"] = cursor
        if limit is not None:
            params["page[limit]"] = limit

        payload = await self._workdrive.request(
            "GET",
            f"/api/v1/files/{folder_id}/recentchanges",
            params=params,
            headers=headers,
        )
        return parse_workdrive_response(payload)



"""Zoho CRM product client."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Protocol

from zoho.core.cache import AsyncTTLCache
from zoho.core.discovery_cache import DiscoveryDiskCache

if TYPE_CHECKING:
    from zoho.crm.discovery import CRMDynamicNamespace
    from zoho.crm.modules import ModulesClient
    from zoho.crm.org import OrgClient
    from zoho.crm.records import RecordsClient
    from zoho.crm.users import UsersClient


class CRMRequestCallable(Protocol):
    async def __call__(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]: ...


class CRMClient:
    """Product-scoped CRM client."""

    def __init__(
        self,
        *,
        request: CRMRequestCallable,
        metadata_cache: AsyncTTLCache[dict[str, Any]] | None,
        default_metadata_ttl_seconds: int,
        discovery_cache: DiscoveryDiskCache | None = None,
        discovery_cache_scope: str = "default:US:production",
    ) -> None:
        self._request_fn = request
        self._metadata_cache = metadata_cache
        self._default_metadata_ttl_seconds = default_metadata_ttl_seconds
        self._discovery_cache = discovery_cache
        self._discovery_cache_scope = discovery_cache_scope

        self._records: RecordsClient | None = None
        self._modules: ModulesClient | None = None
        self._org: OrgClient | None = None
        self._users: UsersClient | None = None
        self._dynamic: CRMDynamicNamespace | None = None

    @property
    def records(self) -> RecordsClient:
        if self._records is None:
            from zoho.crm.records import RecordsClient

            self._records = RecordsClient(self)
        return self._records

    @property
    def modules(self) -> ModulesClient:
        if self._modules is None:
            from zoho.crm.modules import ModulesClient

            self._modules = ModulesClient(self)
        return self._modules

    @property
    def org(self) -> OrgClient:
        if self._org is None:
            from zoho.crm.org import OrgClient

            self._org = OrgClient(self)
        return self._org

    @property
    def users(self) -> UsersClient:
        if self._users is None:
            from zoho.crm.users import UsersClient

            self._users = UsersClient(self)
        return self._users

    @property
    def dynamic(self) -> CRMDynamicNamespace:
        """Access runtime module discovery and dynamic module clients."""

        if self._dynamic is None:
            from zoho.crm.discovery import CRMDynamicNamespace

            self._dynamic = CRMDynamicNamespace(
                self,
                discovery_cache=self._discovery_cache,
                discovery_cache_scope=self._discovery_cache_scope,
            )
        return self._dynamic

    async def request(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
        use_cache: bool = False,
        cache_key: str | None = None,
        cache_ttl_seconds: int | None = None,
    ) -> dict[str, Any]:
        if use_cache and method.upper() == "GET" and self._metadata_cache is not None:
            key = cache_key or self._build_cache_key(method=method, path=path, params=params)
            cached = await self._metadata_cache.get(key)
            if cached is not None:
                return cached

            payload = await self._request_fn(
                method,
                path,
                headers=headers,
                params=params,
                json=json,
                data=data,
                files=files,
                timeout=timeout,
            )
            ttl = (
                cache_ttl_seconds
                if cache_ttl_seconds is not None
                else self._default_metadata_ttl_seconds
            )
            await self._metadata_cache.set(key, payload, ttl)
            return payload

        return await self._request_fn(
            method,
            path,
            headers=headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )

    @staticmethod
    def _build_cache_key(*, method: str, path: str, params: Mapping[str, Any] | None = None) -> str:
        parts = [method.upper(), path]
        if params:
            parts.extend(f"{key}={params[key]}" for key in sorted(params))
        return "|".join(parts)



"""WorkDrive domain models."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class WorkDriveResource(BaseModel):
    """JSON:API-like WorkDrive resource."""

    model_config = ConfigDict(extra="allow")

    id: str | None = None
    type: str | None = None
    attributes: dict[str, Any] = Field(default_factory=dict)
    relationships: dict[str, Any] = Field(default_factory=dict)
    links: dict[str, Any] = Field(default_factory=dict)


class WorkDriveResponse(BaseModel):
    """Common WorkDrive response envelope."""

    model_config = ConfigDict(extra="allow")

    data: WorkDriveResource | list[WorkDriveResource] | None = None
    meta: dict[str, Any] = Field(default_factory=dict)
    links: dict[str, Any] = Field(default_factory=dict)
    errors: list[dict[str, Any]] = Field(default_factory=list)

    @property
    def resources(self) -> list[WorkDriveResource]:
        if isinstance(self.data, list):
            return self.data
        if isinstance(self.data, WorkDriveResource):
            return [self.data]
        return []


def parse_workdrive_response(payload: Mapping[str, Any]) -> WorkDriveResponse:
    """Convert raw WorkDrive payloads into typed response models."""

    return WorkDriveResponse.model_validate(dict(payload))



"""Zoho Projects product client."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    from zoho.projects.portals import ProjectsPortalsClient
    from zoho.projects.projects import ProjectsProjectsClient
    from zoho.projects.tasks import ProjectsTasksClient


class ProjectsRequestCallable(Protocol):
    async def __call__(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]: ...


class ProjectsClient:
    """Product-scoped Projects client (V3 endpoints).

    Example:
        ```python
        projects = await zoho.projects.projects.list(portal_id="12345678")
        print(projects[0].id if projects else None)
        ```
    """

    def __init__(
        self, *, request: ProjectsRequestCallable, default_portal_id: str | None = None
    ) -> None:
        self._request_fn = request
        self._default_portal_id = default_portal_id
        self._portals: ProjectsPortalsClient | None = None
        self._projects: ProjectsProjectsClient | None = None
        self._tasks: ProjectsTasksClient | None = None

    @property
    def portals(self) -> ProjectsPortalsClient:
        if self._portals is None:
            from zoho.projects.portals import ProjectsPortalsClient

            self._portals = ProjectsPortalsClient(self)
        return self._portals

    @property
    def projects(self) -> ProjectsProjectsClient:
        if self._projects is None:
            from zoho.projects.projects import ProjectsProjectsClient

            self._projects = ProjectsProjectsClient(self)
        return self._projects

    @property
    def tasks(self) -> ProjectsTasksClient:
        if self._tasks is None:
            from zoho.projects.tasks import ProjectsTasksClient

            self._tasks = ProjectsTasksClient(self)
        return self._tasks

    def resolve_portal_id(self, portal_id: str | int | None) -> str:
        resolved = str(portal_id) if portal_id is not None else self._default_portal_id
        if resolved is None or not str(resolved).strip():
            raise ValueError(
                "portal_id is required (pass explicitly or configure projects_default_portal_id)"
            )
        return str(resolved)

    async def request(
        self,
        method: str,
        path: str,
        *,
        headers: Mapping[str, str] | None = None,
        params: Mapping[str, Any] | None = None,
        json: Any | None = None,
        data: Any | None = None,
        files: Any | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        return await self._request_fn(
            method,
            path,
            headers=headers,
            params=params,
            json=json,
            data=data,
            files=files,
            timeout=timeout,
        )



"""Zoho WorkDrive namespace."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.workdrive.client import WorkDriveClient

__all__ = ["WorkDriveClient"]


def __getattr__(name: str) -> Any:
    if name == "WorkDriveClient":
        from zoho.workdrive.client import WorkDriveClient

        return WorkDriveClient
    raise AttributeError(f"module 'zoho.workdrive' has no attribute {name!r}")



"""Cliq channels APIs."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any

from zoho.cliq.models import CliqResponse, parse_cliq_response

if TYPE_CHECKING:
    from zoho.cliq.client import CliqClient


class CliqChannelsClient:
    """Channel APIs for Zoho Cliq."""

    def __init__(self, cliq_client: CliqClient) -> None:
        self._cliq = cliq_client

    async def list(
        self,
        *,
        name: str | None = None,
        status: str | None = None,
        level: str | None = None,
        limit: int | None = None,
        next_token: str | None = None,
        modified_before: int | None = None,
        modified_after: int | None = None,
        created_before: int | None = None,
        created_after: int | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        params: dict[str, Any] = {}
        if name:
            params["name"] = name
        if status:
            params["status"] = status
        if level:
            params["level"] = level
        if limit is not None:
            params["limit"] = limit
        if next_token:
            params["next_token"] = next_token
        if modified_before is not None:
            params["modified_before"] = modified_before
        if modified_after is not None:
            params["modified_after"] = modified_after
        if created_before is not None:
            params["created_before"] = created_before
        if created_after is not None:
            params["created_after"] = created_after

        payload = await self._cliq.request("GET", "/channels", params=params, headers=headers)
        return parse_cliq_response(payload)

    async def get(
        self,
        *,
        channel_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request("GET", f"/channels/{channel_id}", headers=headers)
        return parse_cliq_response(payload)

    async def create(
        self,
        *,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request("POST", "/channels", json=dict(data), headers=headers)
        return parse_cliq_response(payload)

    async def update(
        self,
        *,
        channel_id: str,
        data: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "PUT",
            f"/channels/{channel_id}",
            json=dict(data),
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def delete(
        self,
        *,
        channel_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request("DELETE", f"/channels/{channel_id}", headers=headers)
        return parse_cliq_response(payload)

    async def list_members(
        self,
        *,
        channel_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "GET", f"/channels/{channel_id}/members", headers=headers
        )
        return parse_cliq_response(payload)

    async def add_members(
        self,
        *,
        channel_id: str,
        user_ids: Sequence[str],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "POST",
            f"/channels/{channel_id}/members",
            json={"user_ids": list(user_ids)},
            headers=headers,
        )
        return parse_cliq_response(payload)

    async def remove_members(
        self,
        *,
        channel_id: str,
        user_ids: Sequence[str],
        headers: Mapping[str, str] | None = None,
    ) -> CliqResponse:
        payload = await self._cliq.request(
            "DELETE",
            f"/channels/{channel_id}/members",
            json={"user_ids": list(user_ids)},
            headers=headers,
        )
        return parse_cliq_response(payload)



"""Projects project operations."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.projects.models import Project, ProjectsResponse, parse_projects_response

if TYPE_CHECKING:
    from zoho.projects.client import ProjectsClient


class ProjectsProjectsClient:
    """Project resource operations under a portal.

    Example:
        ```python
        created = await zoho.projects.projects.create(
            portal_id="12345678",
            data={"name": "SDK Rollout"},
        )
        print(created.status)
        ```
    """

    def __init__(self, projects_client: ProjectsClient) -> None:
        self._projects = projects_client

    async def list(
        self,
        *,
        portal_id: str | int | None = None,
        page: int = 1,
        per_page: int = 200,
    ) -> list[Project]:
        resolved = self._projects.resolve_portal_id(portal_id)
        payload = await self._projects.request(
            "GET",
            f"/portal/{resolved}/projects",
            params={"page": page, "per_page": per_page},
        )
        return parse_projects_response(payload).projects

    async def get(
        self, *, project_id: str | int, portal_id: str | int | None = None
    ) -> Project | None:
        resolved = self._projects.resolve_portal_id(portal_id)
        payload = await self._projects.request("GET", f"/portal/{resolved}/projects/{project_id}")
        response = parse_projects_response(payload)

        if response.project is not None:
            return response.project
        if response.projects:
            return response.projects[0]
        return None

    async def create(
        self,
        *,
        data: Mapping[str, Any],
        portal_id: str | int | None = None,
    ) -> ProjectsResponse:
        resolved = self._projects.resolve_portal_id(portal_id)
        payload = await self._projects.request(
            "POST",
            f"/portal/{resolved}/projects",
            json=dict(data),
        )
        return parse_projects_response(payload)

    async def update(
        self,
        *,
        project_id: str | int,
        data: Mapping[str, Any],
        portal_id: str | int | None = None,
    ) -> ProjectsResponse:
        resolved = self._projects.resolve_portal_id(portal_id)
        payload = await self._projects.request(
            "PATCH",
            f"/portal/{resolved}/projects/{project_id}",
            json=dict(data),
        )
        return parse_projects_response(payload)

    async def trash(
        self, *, project_id: str | int, portal_id: str | int | None = None
    ) -> ProjectsResponse:
        resolved = self._projects.resolve_portal_id(portal_id)
        payload = await self._projects.request(
            "POST", f"/portal/{resolved}/projects/{project_id}/trash"
        )
        return parse_projects_response(payload)



"""Mail domain models."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel, ConfigDict


class MailResponse(BaseModel):
    """Common Mail response envelope."""

    model_config = ConfigDict(extra="allow")

    status: dict[str, Any] | str | None = None
    data: dict[str, Any] | list[dict[str, Any]] | None = None

    @property
    def result_rows(self) -> list[dict[str, Any]]:
        if isinstance(self.data, list):
            return [item for item in self.data if isinstance(item, dict)]
        if isinstance(self.data, dict):
            for key in (
                "accounts",
                "folders",
                "messages",
                "threads",
                "data",
            ):
                value = self.data.get(key)
                if isinstance(value, list):
                    return [item for item in value if isinstance(item, dict)]
        return []


def parse_mail_response(payload: Mapping[str, Any]) -> MailResponse:
    """Convert raw Mail payloads into typed response models."""

    return MailResponse.model_validate(dict(payload))



"""Mail accounts APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from zoho.mail.models import MailResponse, parse_mail_response

if TYPE_CHECKING:
    from zoho.mail.client import MailClient


class MailAccountsClient:
    """Account APIs for Zoho Mail."""

    def __init__(self, mail_client: MailClient) -> None:
        self._mail = mail_client

    async def list(self, *, headers: Mapping[str, str] | None = None) -> MailResponse:
        payload = await self._mail.request("GET", "/accounts", headers=headers)
        return parse_mail_response(payload)

    async def get(
        self,
        *,
        account_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        payload = await self._mail.request("GET", f"/accounts/{account_id}", headers=headers)
        return parse_mail_response(payload)



"""Ingestion iterator helpers for Zoho Sheet."""

from __future__ import annotations

from collections.abc import AsyncIterator
from typing import Any

from zoho.client import Zoho
from zoho.ingestion._common import resolve_connection
from zoho.ingestion.models import IngestionBatch, IngestionCheckpoint, IngestionDocument


def _extract_rows(payload: dict[str, Any]) -> list[dict[str, Any]]:
    rows = payload.get("records")
    if isinstance(rows, list):
        return [item for item in rows if isinstance(item, dict)]

    data = payload.get("data")
    if isinstance(data, list):
        return [item for item in data if isinstance(item, dict)]
    return []


async def iter_sheet_worksheet_documents(
    client: Zoho,
    *,
    workbook_id: str,
    worksheet_name: str,
    connection_name: str = "default",
    page_size: int = 200,
    checkpoint: IngestionCheckpoint | None = None,
    max_pages: int | None = None,
) -> AsyncIterator[IngestionBatch]:
    """Yield normalized document batches from Sheet worksheet records."""

    target = resolve_connection(client, connection_name)
    offset = checkpoint.offset if checkpoint and checkpoint.offset is not None else 0

    pages = 0
    while True:
        response = await target.sheet.tabular.fetch_worksheet_records(
            workbook_id=workbook_id,
            worksheet_name=worksheet_name,
            offset=offset,
            limit=page_size,
        )
        rows = _extract_rows(response.model_dump(mode="python"))
        if not rows:
            break

        docs: list[IngestionDocument] = []
        for row in rows:
            rid = row.get("id") or row.get("row_id") or f"{worksheet_name}:{offset}:{len(docs)}"
            docs.append(
                IngestionDocument(
                    id=str(rid),
                    source="zoho.sheet",
                    title=str(row.get("title") or row.get("name") or rid),
                    content=str(row.get("content") or row.get("value") or "") or None,
                    updated_at=(
                        str(row.get("modified_time") or row.get("updated_at"))
                        if row.get("modified_time") or row.get("updated_at")
                        else None
                    ),
                    metadata={"workbook_id": workbook_id, "worksheet_name": worksheet_name},
                    raw=row,
                )
            )

        next_offset = offset + len(rows)
        yield IngestionBatch(
            source="zoho.sheet",
            documents=docs,
            checkpoint=IngestionCheckpoint(
                offset=next_offset,
                extras={"workbook_id": workbook_id, "worksheet_name": worksheet_name},
            ),
        )

        pages += 1
        if len(rows) < page_size:
            break
        if max_pages is not None and pages >= max_pages:
            break
        offset = next_offset



"""Ingestion iterator helpers for Zoho People."""

from __future__ import annotations

from collections.abc import AsyncIterator

from zoho.client import Zoho
from zoho.ingestion._common import resolve_connection
from zoho.ingestion.models import IngestionBatch, IngestionCheckpoint, IngestionDocument


async def iter_people_form_documents(
    client: Zoho,
    *,
    form_link_name: str,
    connection_name: str = "default",
    page_size: int = 200,
    checkpoint: IngestionCheckpoint | None = None,
    max_pages: int | None = None,
) -> AsyncIterator[IngestionBatch]:
    """Yield normalized document batches from People form records."""

    target = resolve_connection(client, connection_name)
    start_index = checkpoint.offset if checkpoint and checkpoint.offset is not None else 0

    pages = 0
    while True:
        response = await target.people.forms.list_records(
            form_link_name=form_link_name,
            from_index=start_index,
            limit=page_size,
        )
        rows = response.result_rows
        if not rows:
            break

        docs: list[IngestionDocument] = []
        for row in rows:
            rid = (
                row.get("recordId")
                or row.get("record_id")
                or row.get("id")
                or row.get("employeeId")
                or row.get("zoho_id")
                or f"{form_link_name}:{start_index}:{len(docs)}"
            )
            docs.append(
                IngestionDocument(
                    id=str(rid),
                    source="zoho.people",
                    title=str(row.get("displayName") or row.get("name") or rid),
                    content=str(row.get("description") or "") or None,
                    updated_at=(
                        str(row.get("modifiedTime") or row.get("modified_time"))
                        if row.get("modifiedTime") or row.get("modified_time")
                        else None
                    ),
                    metadata={"form_link_name": form_link_name},
                    raw=row,
                )
            )

        next_offset = start_index + len(rows)
        yield IngestionBatch(
            source="zoho.people",
            documents=docs,
            checkpoint=IngestionCheckpoint(
                offset=next_offset, extras={"form_link_name": form_link_name}
            ),
        )

        pages += 1
        if len(rows) < page_size:
            break
        if max_pages is not None and pages >= max_pages:
            break
        start_index = next_offset



"""Zoho People namespace."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.people.client import PeopleClient

__all__ = ["PeopleClient"]


def __getattr__(name: str) -> Any:
    if name == "PeopleClient":
        from zoho.people.client import PeopleClient

        return PeopleClient
    raise AttributeError(f"module 'zoho.people' has no attribute {name!r}")



"""OAuth2 refresh-token auth provider."""

from __future__ import annotations

from collections.abc import Mapping
from datetime import UTC, datetime, timedelta
from typing import Any

import httpx

from zoho.core.errors import ZohoAuthError
from zoho.core.logging import get_logger
from zoho.core.token_store import OAuthToken, TokenStore

_ACCOUNTS_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://accounts.zoho.com",
    "EU": "https://accounts.zoho.eu",
    "IN": "https://accounts.zoho.in",
    "AU": "https://accounts.zoho.com.au",
    "JP": "https://accounts.zoho.jp",
    "CA": "https://accounts.zohocloud.ca",
    "SA": "https://accounts.zoho.sa",
    "CN": "https://accounts.zoho.com.cn",
}

_API_DOMAIN_BY_DC: dict[str, str] = {
    "US": "https://www.zohoapis.com",
    "EU": "https://www.zohoapis.eu",
    "IN": "https://www.zohoapis.in",
    "AU": "https://www.zohoapis.com.au",
    "JP": "https://www.zohoapis.jp",
    "CA": "https://www.zohocloud.ca",
    "SA": "https://www.zohoapis.sa",
    "CN": "https://www.zohoapis.com.cn",
}


class OAuth2RefreshAuthProvider:
    """Refresh-token OAuth provider with persisted token caching."""

    def __init__(
        self,
        *,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        dc: str,
        environment: str,
        token_store: TokenStore,
        accounts_domain: str | None = None,
        api_domain: str | None = None,
        cache_namespace: str = "default",
        user_agent: str = "zoho-sdk/0.1.3",
        http_client: httpx.AsyncClient | None = None,
    ) -> None:
        self._client_id = client_id
        self._client_secret = client_secret
        self._refresh_token = refresh_token
        self._dc = dc
        self._environment = environment
        self._token_store = token_store
        self._accounts_domain = (accounts_domain or _ACCOUNTS_DOMAIN_BY_DC[dc]).rstrip("/")
        self._api_domain_override = api_domain
        self._user_agent = user_agent
        self._logger = get_logger("zoho.auth")

        self._http_client = http_client or httpx.AsyncClient(timeout=20.0)
        self._owns_http_client = http_client is None

        self._cache_key = f"{cache_namespace}:{client_id}:{dc}:{environment}:default"

    async def close(self) -> None:
        if self._owns_http_client:
            await self._http_client.aclose()

    async def get_auth_headers(self) -> dict[str, str]:
        token = await self.get_token()
        return {"Authorization": f"Zoho-oauthtoken {token.access_token}"}

    async def get_api_domain(self) -> str:
        if self._api_domain_override:
            return self._api_domain_override.rstrip("/")
        token = await self.get_token()
        if token.api_domain:
            return token.api_domain.rstrip("/")
        return _API_DOMAIN_BY_DC[self._dc]

    async def get_token(self) -> OAuthToken:
        existing = await self._token_store.load(self._cache_key)
        if existing and not existing.is_expired():
            return existing

        async with self._token_store.refresh_lock(self._cache_key):
            existing = await self._token_store.load(self._cache_key)
            if existing and not existing.is_expired():
                return existing

            refreshed = await self._refresh(existing)
            await self._token_store.save(self._cache_key, refreshed)
            return refreshed

    async def _refresh(self, existing: OAuthToken | None) -> OAuthToken:
        token_url = f"{self._accounts_domain}/oauth/v2/token"
        refresh_token = (
            existing.refresh_token
            if existing is not None and existing.refresh_token
            else self._refresh_token
        )

        payload = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self._client_id,
            "client_secret": self._client_secret,
        }

        response = await self._http_client.post(
            token_url,
            data=payload,
            headers={"User-Agent": self._user_agent},
        )

        body: Mapping[str, Any]
        try:
            body = response.json()
        except ValueError:
            body = {}

        if response.status_code >= 400:
            raise ZohoAuthError(
                body.get("message", "Failed to refresh access token")
                if isinstance(body, Mapping)
                else "Failed to refresh access token",
                status_code=response.status_code,
                code=body.get("code") if isinstance(body, Mapping) else None,
                details=body if isinstance(body, Mapping) else None,
                method="POST",
                url=token_url,
                response_text=response.text[:500],
            )

        access_token = body.get("access_token") if isinstance(body, Mapping) else None
        if not isinstance(access_token, str) or not access_token:
            raise ZohoAuthError(
                "Token response did not include access_token",
                status_code=response.status_code,
                details=body if isinstance(body, Mapping) else None,
                method="POST",
                url=token_url,
            )

        expires_in_raw = body.get("expires_in_sec") if isinstance(body, Mapping) else None
        if expires_in_raw is None and isinstance(body, Mapping):
            expires_in_raw = body.get("expires_in")

        expires_in = int(expires_in_raw) if expires_in_raw is not None else 3600
        api_domain = body.get("api_domain") if isinstance(body, Mapping) else None
        scope_raw = body.get("scope") if isinstance(body, Mapping) else None
        scopes = tuple(str(scope_raw).split(",")) if scope_raw else ()

        token = OAuthToken(
            access_token=access_token,
            refresh_token=refresh_token,
            expires_at=datetime.now(UTC) + timedelta(seconds=max(expires_in, 1)),
            api_domain=(
                str(api_domain) if api_domain else existing.api_domain if existing else None
            ),
            scope=scopes,
        )

        self._logger.info("oauth_token_refreshed", dc=self._dc, environment=self._environment)
        return token



"""Analytics bulk/import-export APIs."""

from __future__ import annotations

import json
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.analytics.models import AnalyticsResponse, parse_analytics_response

if TYPE_CHECKING:
    from zoho.analytics.client import AnalyticsClient


def _config_params(config: Mapping[str, Any] | None) -> dict[str, Any]:
    if not config:
        return {}
    return {"CONFIG": json.dumps(dict(config), separators=(",", ":"), ensure_ascii=True)}


class AnalyticsBulkClient:
    """Bulk and asynchronous job APIs for Zoho Analytics."""

    def __init__(self, analytics_client: AnalyticsClient) -> None:
        self._analytics = analytics_client

    async def import_data(
        self,
        *,
        workspace_id: str,
        config: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> AnalyticsResponse:
        payload = await self._analytics.request(
            "POST",
            f"/bulk/workspaces/{workspace_id}/data",
            params=_config_params(config),
            headers=headers,
        )
        return parse_analytics_response(payload)

    async def export_data(
        self,
        *,
        workspace_id: str,
        view_id: str,
        config: Mapping[str, Any],
        headers: Mapping[str, str] | None = None,
    ) -> AnalyticsResponse:
        payload = await self._analytics.request(
            "GET",
            f"/bulk/workspaces/{workspace_id}/views/{view_id}/data",
            params=_config_params(config),
            headers=headers,
        )
        return parse_analytics_response(payload)

    async def get_import_job(
        self,
        *,
        workspace_id: str,
        job_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> AnalyticsResponse:
        payload = await self._analytics.request(
            "GET",
            f"/bulk/workspaces/{workspace_id}/importjobs/{job_id}",
            headers=headers,
        )
        return parse_analytics_response(payload)

    async def get_export_job(
        self,
        *,
        workspace_id: str,
        job_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> AnalyticsResponse:
        payload = await self._analytics.request(
            "GET",
            f"/bulk/workspaces/{workspace_id}/exportjobs/{job_id}",
            headers=headers,
        )
        return parse_analytics_response(payload)

    async def download_export_job(
        self,
        *,
        workspace_id: str,
        job_id: str,
        headers: Mapping[str, str] | None = None,
    ) -> AnalyticsResponse:
        payload = await self._analytics.request(
            "GET",
            f"/bulk/workspaces/{workspace_id}/exportjobs/{job_id}/data",
            headers=headers,
        )
        return parse_analytics_response(payload)



"""Zoho Writer namespace."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.writer.client import WriterClient

__all__ = ["WriterClient"]


def __getattr__(name: str) -> Any:
    if name == "WriterClient":
        from zoho.writer.client import WriterClient

        return WriterClient
    raise AttributeError(f"module 'zoho.writer' has no attribute {name!r}")



"""Ingestion-oriented iterators and checkpoint helpers."""

from zoho.ingestion.analytics import (
    iter_analytics_view_documents,
    iter_analytics_workspace_documents,
)
from zoho.ingestion.cliq import (
    iter_cliq_channel_documents,
    iter_cliq_chat_documents,
    iter_cliq_thread_documents,
)
from zoho.ingestion.crm import iter_crm_documents, iter_crm_module_documents
from zoho.ingestion.mail import iter_mail_message_documents
from zoho.ingestion.models import IngestionBatch, IngestionCheckpoint, IngestionDocument
from zoho.ingestion.people import iter_people_form_documents
from zoho.ingestion.sheet import iter_sheet_worksheet_documents
from zoho.ingestion.workdrive import iter_workdrive_recent_documents
from zoho.ingestion.writer import iter_writer_document_documents

__all__ = [
    "IngestionBatch",
    "IngestionCheckpoint",
    "IngestionDocument",
    "iter_analytics_view_documents",
    "iter_analytics_workspace_documents",
    "iter_cliq_channel_documents",
    "iter_cliq_chat_documents",
    "iter_cliq_thread_documents",
    "iter_crm_documents",
    "iter_crm_module_documents",
    "iter_mail_message_documents",
    "iter_people_form_documents",
    "iter_sheet_worksheet_documents",
    "iter_workdrive_recent_documents",
    "iter_writer_document_documents",
]



"""Mail folders APIs."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from zoho.mail.models import MailResponse, parse_mail_response

if TYPE_CHECKING:
    from zoho.mail.client import MailClient


class MailFoldersClient:
    """Folder APIs for Zoho Mail."""

    def __init__(self, mail_client: MailClient) -> None:
        self._mail = mail_client

    async def list(
        self,
        *,
        account_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        payload = await self._mail.request(
            "GET", f"/accounts/{account_id}/folders", headers=headers
        )
        return parse_mail_response(payload)

    async def get(
        self,
        *,
        account_id: str | int,
        folder_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        payload = await self._mail.request(
            "GET",
            f"/accounts/{account_id}/folders/{folder_id}",
            headers=headers,
        )
        return parse_mail_response(payload)

    async def create(
        self,
        *,
        account_id: str | int,
        folder_name: str,
        parent_folder_id: str | int | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        data: dict[str, Any] = {"folderName": folder_name}
        if parent_folder_id is not None:
            data["parentFolderId"] = str(parent_folder_id)

        payload = await self._mail.request(
            "POST",
            f"/accounts/{account_id}/folders",
            json=data,
            headers=headers,
        )
        return parse_mail_response(payload)

    async def update(
        self,
        *,
        account_id: str | int,
        folder_id: str | int,
        folder_name: str,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        payload = await self._mail.request(
            "PUT",
            f"/accounts/{account_id}/folders/{folder_id}",
            json={"folderName": folder_name},
            headers=headers,
        )
        return parse_mail_response(payload)

    async def delete(
        self,
        *,
        account_id: str | int,
        folder_id: str | int,
        headers: Mapping[str, str] | None = None,
    ) -> MailResponse:
        payload = await self._mail.request(
            "DELETE",
            f"/accounts/{account_id}/folders/{folder_id}",
            headers=headers,
        )
        return parse_mail_response(payload)



"""Token persistence backends for OAuth refresh tokens."""

from __future__ import annotations

import asyncio
import contextlib
import inspect
import sqlite3
import time
import uuid
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import UTC, datetime, timedelta
from pathlib import Path
from typing import Any, Protocol, cast

from pydantic import BaseModel, Field

redis_asyncio: Any
try:
    import redis.asyncio as redis_asyncio
except ImportError:  # pragma: no cover - optional dependency
    redis_asyncio = None


class OAuthToken(BaseModel):
    """Persisted OAuth token representation."""

    access_token: str
    expires_at: datetime
    refresh_token: str | None = None
    api_domain: str | None = None
    scope: tuple[str, ...] = Field(default_factory=tuple)
    token_type: str = "Zoho-oauthtoken"
    created_at: datetime = Field(default_factory=lambda: datetime.now(UTC))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(UTC))

    def is_expired(self, *, skew_seconds: int = 30) -> bool:
        return self.expires_at <= datetime.now(UTC) + timedelta(seconds=skew_seconds)


class TokenStore(Protocol):
    """Token store protocol used by auth providers."""

    async def load(self, key: str) -> OAuthToken | None: ...

    async def save(self, key: str, token: OAuthToken) -> None: ...

    def refresh_lock(
        self, key: str, *, ttl_seconds: int = 30, wait_timeout_seconds: float = 10.0
    ) -> contextlib.AbstractAsyncContextManager[None]: ...


class MemoryTokenStore:
    """In-memory token store for tests and ephemeral processes."""

    def __init__(self) -> None:
        self._tokens: dict[str, OAuthToken] = {}
        self._locks: dict[str, asyncio.Lock] = {}

    async def load(self, key: str) -> OAuthToken | None:
        return self._tokens.get(key)

    async def save(self, key: str, token: OAuthToken) -> None:
        token.updated_at = datetime.now(UTC)
        self._tokens[key] = token

    def _get_lock(self, key: str) -> asyncio.Lock:
        lock = self._locks.get(key)
        if lock is None:
            lock = asyncio.Lock()
            self._locks[key] = lock
        return lock

    @asynccontextmanager
    async def refresh_lock(
        self, key: str, *, ttl_seconds: int = 30, wait_timeout_seconds: float = 10.0
    ) -> AsyncIterator[None]:
        _ = ttl_seconds
        _ = wait_timeout_seconds
        lock = self._get_lock(key)
        await lock.acquire()
        try:
            yield
        finally:
            lock.release()


class SQLiteTokenStore:
    """SQLite-backed token store with cross-process refresh locking."""

    def __init__(self, path: Path) -> None:
        self._path = path.expanduser()
        self._initialized = False
        self._init_lock = asyncio.Lock()

    async def _ensure_initialized(self) -> None:
        if self._initialized:
            return
        async with self._init_lock:
            if self._initialized:
                return
            await asyncio.to_thread(self._initialize_db)
            self._initialized = True

    def _connect(self) -> sqlite3.Connection:
        return sqlite3.connect(str(self._path), timeout=30)

    def _initialize_db(self) -> None:
        self._path.parent.mkdir(parents=True, exist_ok=True)
        with self._connect() as conn:
            conn.execute("PRAGMA journal_mode=WAL")
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS tokens (
                    key TEXT PRIMARY KEY,
                    value TEXT NOT NULL,
                    expires_at INTEGER NOT NULL,
                    updated_at INTEGER NOT NULL
                )
                """
            )
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS locks (
                    key TEXT PRIMARY KEY,
                    owner TEXT NOT NULL,
                    expires_at INTEGER NOT NULL
                )
                """
            )
            conn.commit()

    async def load(self, key: str) -> OAuthToken | None:
        await self._ensure_initialized()
        return await asyncio.to_thread(self._load_sync, key)

    def _load_sync(self, key: str) -> OAuthToken | None:
        with self._connect() as conn:
            row = conn.execute("SELECT value FROM tokens WHERE key = ?", (key,)).fetchone()
        if row is None:
            return None
        return OAuthToken.model_validate_json(row[0])

    async def save(self, key: str, token: OAuthToken) -> None:
        await self._ensure_initialized()
        token.updated_at = datetime.now(UTC)
        await asyncio.to_thread(self._save_sync, key, token)

    def _save_sync(self, key: str, token: OAuthToken) -> None:
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO tokens(key, value, expires_at, updated_at)
                VALUES (?, ?, ?, ?)
                ON CONFLICT(key) DO UPDATE SET
                    value=excluded.value,
                    expires_at=excluded.expires_at,
                    updated_at=excluded.updated_at
                """,
                (
                    key,
                    token.model_dump_json(),
                    int(token.expires_at.timestamp()),
                    int(time.time()),
                ),
            )
            conn.commit()

    def _try_acquire_lock(self, key: str, owner: str, ttl_seconds: int) -> bool:
        now = int(time.time())
        expires_at = now + ttl_seconds
        with self._connect() as conn:
            conn.execute("BEGIN IMMEDIATE")
            conn.execute("DELETE FROM locks WHERE key = ? AND expires_at <= ?", (key, now))
            try:
                conn.execute(
                    "INSERT INTO locks(key, owner, expires_at) VALUES (?, ?, ?)",
                    (key, owner, expires_at),
                )
            except sqlite3.IntegrityError:
                conn.rollback()
                return False
            conn.commit()
            return True

    def _release_lock(self, key: str, owner: str) -> None:
        with self._connect() as conn:
            conn.execute("DELETE FROM locks WHERE key = ? AND owner = ?", (key, owner))
            conn.commit()

    @asynccontextmanager
    async def refresh_lock(
        self, key: str, *, ttl_seconds: int = 30, wait_timeout_seconds: float = 10.0
    ) -> AsyncIterator[None]:
        await self._ensure_initialized()

        owner = uuid.uuid4().hex
        acquired = False
        deadline = time.monotonic() + wait_timeout_seconds

        while time.monotonic() < deadline:
            acquired = await asyncio.to_thread(self._try_acquire_lock, key, owner, ttl_seconds)
            if acquired:
                break
            await asyncio.sleep(0.1)

        if not acquired:
            raise TimeoutError(f"Unable to acquire token refresh lock for key={key!r}")

        try:
            yield
        finally:
            await asyncio.to_thread(self._release_lock, key, owner)


class RedisTokenStore:
    """Redis-backed token store for distributed worker deployments."""

    def __init__(self, redis_url: str) -> None:
        if redis_asyncio is None:  # pragma: no cover - optional dependency
            raise ImportError("redis extra is required for RedisTokenStore")
        self._client = redis_asyncio.from_url(redis_url, decode_responses=True)

    @staticmethod
    def _token_key(key: str) -> str:
        return f"zoho:token:{key}"

    @staticmethod
    def _lock_key(key: str) -> str:
        return f"zoho:lock:{key}"

    async def load(self, key: str) -> OAuthToken | None:
        raw = await self._client.get(self._token_key(key))
        if raw is None:
            return None
        return OAuthToken.model_validate_json(raw)

    async def save(self, key: str, token: OAuthToken) -> None:
        token.updated_at = datetime.now(UTC)
        ttl_seconds = max(int((token.expires_at - datetime.now(UTC)).total_seconds()), 1)
        await self._client.set(self._token_key(key), token.model_dump_json(), ex=ttl_seconds)

    @asynccontextmanager
    async def refresh_lock(
        self, key: str, *, ttl_seconds: int = 30, wait_timeout_seconds: float = 10.0
    ) -> AsyncIterator[None]:
        lock_key = self._lock_key(key)
        owner = uuid.uuid4().hex
        acquired = False

        deadline = time.monotonic() + wait_timeout_seconds
        while time.monotonic() < deadline:
            acquired = bool(await self._client.set(lock_key, owner, nx=True, ex=ttl_seconds))
            if acquired:
                break
            await asyncio.sleep(0.1)

        if not acquired:
            raise TimeoutError(f"Unable to acquire token refresh lock for key={key!r}")

        try:
            yield
        finally:
            # Compare-and-delete to avoid removing someone else's lock.
            script = (
                "if redis.call('GET', KEYS[1]) == ARGV[1] then "
                "return redis.call('DEL', KEYS[1]) else return 0 end"
            )
            eval_result = self._client.eval(script, 1, lock_key, owner)
            if inspect.isawaitable(eval_result):
                await cast(Any, eval_result)


def build_token_store(*, backend: str, sqlite_path: Path, redis_url: str | None) -> TokenStore:
    """Create a concrete token store backend from settings."""

    if backend == "memory":
        return MemoryTokenStore()
    if backend == "sqlite":
        return SQLiteTokenStore(sqlite_path)
    if backend == "redis":
        if not redis_url:
            raise ValueError("redis_url is required when token_store_backend='redis'")
        return RedisTokenStore(redis_url)
    raise ValueError(f"Unsupported token store backend: {backend}")



"""Zoho CRM namespace."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from zoho.crm.client import CRMClient
    from zoho.crm.models import Record

__all__ = ["CRMClient", "Record"]


def __getattr__(name: str) -> Any:
    if name == "CRMClient":
        from zoho.crm.client import CRMClient

        return CRMClient
    if name == "Record":
        from zoho.crm.models import Record

        return Record
    raise AttributeError(f"module 'zoho.crm' has no attribute {name!r}")



"""CRM domain models."""

from __future__ import annotations

from collections.abc import Iterator, Mapping, MutableMapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class PageInfo(BaseModel):
    """Pagination details returned by list endpoints."""

    per_page: int | None = None
    count: int | None = None
    page: int | None = None
    more_records: bool = False
    next_page_token: str | None = None


class RecordListResponse(BaseModel):
    """Structured record list response."""

    data: list[dict[str, Any]] = Field(default_factory=list)
    info: PageInfo | None = None


class ActionResponse(BaseModel):
    """Action response envelope used by create/update/delete operations."""

    model_config = ConfigDict(extra="allow")

    status: str | None = None
    code: str | None = None
    message: str | None = None
    details: dict[str, Any] = Field(default_factory=dict)


class Record(MutableMapping[str, Any]):
    """Ergonomic dynamic CRM record mapping.

    Example:
        ```python
        record = Record({"Last_Name": "Ng", "Email": "jane@example.com"})
        record["Company"] = "Acme"
        payload = record.to_dict()
        ```
    """

    def __init__(self, initial: Mapping[str, Any] | None = None) -> None:
        self._values: dict[str, Any] = dict(initial or {})

    @property
    def id(self) -> str | None:
        value = self._values.get("id")
        return str(value) if value is not None else None

    def to_dict(self) -> dict[str, Any]:
        return dict(self._values)

    def __getitem__(self, key: str) -> Any:
        return self._values[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._values[key] = value

    def __delitem__(self, key: str) -> None:
        del self._values[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._values)

    def __len__(self) -> int:
        return len(self._values)


def extract_first_data(payload: Mapping[str, Any]) -> dict[str, Any]:
    data = payload.get("data")
    if not isinstance(data, list) or not data:
        return {}
    first = data[0]
    return first if isinstance(first, dict) else {}



"""Sheet domain models."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class SheetResponse(BaseModel):
    """Common Sheet response envelope."""

    model_config = ConfigDict(extra="allow")

    status: str | None = None
    message: str | None = None
    data: dict[str, Any] | list[dict[str, Any]] | None = None
    workbooks: list[dict[str, Any]] = Field(default_factory=list)
    worksheets: list[dict[str, Any]] = Field(default_factory=list)
    records: list[dict[str, Any]] = Field(default_factory=list)


def parse_sheet_response(payload: Mapping[str, Any]) -> SheetResponse:
    """Convert raw Sheet payloads into typed response models."""

    return SheetResponse.model_validate(dict(payload))
